'''Tests for discoverer_auto.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
from pathlib import Path

import pytest

from ...communicators.stub_communicator import CommunicatorStub
from ...config_components.distributed_config import DistributedConfig
from ...wrangler.constants import PACKAGE_NAME

from ..static_discoverer import StaticDiscoverer
from .discoverer import Discoverer
from .discoverer_auto import DiscovererCatalogAuto

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code


FAKE_DISCOVERER_SOURCE = '''
from typing import Any, List, Optional
from ngautonml.discoverers.impl.discoverer import Discoverer
from ngautonml.neighbor_manager.event import Event
from ngautonml.discoverers.impl.discoverer_catalog import DiscovererCatalog
from ngautonml.neighbor_manager.node_id import NodeID


class FakeDiscoverer(Discoverer):
    name = 'fake_discoverer'
    tags = {
        'some_tag': ['some_value']
    }

    def start(self):
        return

    def laplacian(self):
        return []


def register(catalog: DiscovererCatalog):
    catalog.register(FakeDiscoverer)
'''


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = (retval / f'.{PACKAGE_NAME}' / 'plugins'
                   / 'discoverers' / 'fake_discoverer.py')
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(FAKE_DISCOVERER_SOURCE)
    return retval


def test_discoverer_catalog_auto() -> None:
    config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '0': [],
                }
            }
        }
    })
    dut = DiscovererCatalogAuto()

    got = dut.lookup_by_name('static')(
        config=config, communicator=CommunicatorStub()
    )
    assert isinstance(got, StaticDiscoverer)


def test_home_directory_auto(tmp_path) -> None:
    os.environ['HOME'] = str(tmp_path)
    config = DistributedConfig(clause={
        'static': {
            'adjacency': {
                '0': [],
            }
        }
    })

    dut = DiscovererCatalogAuto()
    got = dut.lookup_by_name('fake_discoverer')(
        config=config, communicator=CommunicatorStub()
    )
    assert isinstance(got, Discoverer)
    assert got.tags['some_tag'] == ['some_value']



'''Catalog for models'''
import abc
from typing import Any, Dict, Optional

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .binder import BinderFactory
from ...catalog.catalog import Catalog
from ...catalog.memory_catalog import MemoryCatalog
from ...catalog.catalog_element_mixin import CatalogElementMixin
from ...communicators.impl.communicator_auto import CommunicatorCatalogAuto
from ...communicators.impl.communicator_catalog import CommunicatorCatalog
from ...discoverers.impl.discoverer_auto import DiscovererCatalogAuto
from ...discoverers.impl.discoverer_catalog import DiscovererCatalog
from ...searcher.param_range import ParamRange

# Things that go in catalogs (like Models) typically have a name and a
# dictionary of tags. Consider extracting this pattern as a mixin.
# We could then remove the following pylint disable.
# pylint: disable=duplicate-code


class Error(BaseException):
    '''Base class for all model-related exceptions.'''


class HyperparamError(Error):
    '''A hyperparam has an unknown name or incorrect value'''


class InputKeyError(Error, KeyError):
    '''Raise when a fit() or predict() method is passed an input dataset
    that is missing required keys.'''


class InputValueError(Error, ValueError):
    '''Raise when a fit() or predict() method is passed an input dataset
    with the correct keys but invalid value(s).'''


class Algorithm(CatalogElementMixin, metaclass=abc.ABCMeta):
    '''This object holds an algorithm.

    This can be an algorithm for training a model, a preprocessing
    algorithm, or a complete trained model.

    .. rubric:: Parameters:

    * ``communincator_catalog``: ``Optional[CommunicatorCatalog]``.
      Catalog that holds communicators for distributed algorithms,
      unused otherwise.
    * ``discoverer_catalog``: ``Optional[DiscovererCatalog]``.
      Catalog that holds discoverers for distributed algorithms,
      unused otherwise.
    * ``**hyperparams``: ``Any``.
      Arbitrary hyperparams that are interpreted by each individual algorithm.

    '''
    _instance_constructor: type  # This is really Type[AlgorithmInstance]
    _name: str = "unnamed"
    _default_hyperparams: Optional[Dict[str, Any]] = None
    _basename: str = ''

    # These two catalogs are only used by distributed algorithms
    _communicator_catalog: CommunicatorCatalog
    _discoverer_catalog: DiscovererCatalog

    # Map from json representation to python representation for specifically named hyperparams.
    # Both json and python elements need to be comparable with ==.
    # Format: {hyperparam_name: {json_value: python_value}}
    _hyperparam_lookup: Dict[str, Dict[Any, Any]]

    def __init__(self,
                 name: Optional[str] = None,
                 communicator_catalog: Optional[CommunicatorCatalog] = None,
                 discoverer_catalog: Optional[DiscovererCatalog] = None,
                 **hyperparams):
        self._name = name or self._name
        self._communicator_catalog = (communicator_catalog or CommunicatorCatalogAuto())
        self._discoverer_catalog = (discoverer_catalog or DiscovererCatalogAuto())
        super().__init__()

        # Convert class _default_hyperparams to an instance-specific
        # data structure.
        if self._default_hyperparams is None:
            self._default_hyperparams = {}
        params = self._default_hyperparams.copy()
        params.update(hyperparams)
        self._default_hyperparams = params
        if not hasattr(self, '_hyperparam_lookup'):
            self._hyperparam_lookup = {}

    @property
    def basename(self):
        '''Return the base name of the model.

        This should be the name used by nf.NeuralForecast to
        label the output column.
        '''
        return self._basename

    @property
    def communicator_catalog(self) -> CommunicatorCatalog:
        '''The communicator catalog'''
        return self._communicator_catalog

    @property
    def discoverer_catalog(self) -> DiscovererCatalog:
        '''The discoverer catalog'''
        return self._discoverer_catalog

    def instantiate(self, **hyperparams) -> Any:
        '''Create an instance of an algorithm that can hold training data.

        Args:
          serialized_model is a saved form of a trained model that instantiate
          can restore.
          hyperparams includes hyperparameters and other parameters needed
          when setting up an instance of the algorithm.

        Returns:
          AlgorithmInstance (marked as Any to break circularity)
        '''
        return self._instance_constructor(
            parent=self,
            **self._default_param_ranges(**self.hyperparams(**hyperparams)))

    def _default_param_ranges(self, **hyperparams) -> Dict[str, Any]:
        '''If any hyperparams are of type ParamRange, replace them with their default values.'''
        return {
            k: (v.default if isinstance(v, ParamRange) else v)
            for k, v in hyperparams.items()
        }

    def hyperparams(self, **overrides) -> Dict[str, Any]:
        '''Get the hyperparameters for this algorithm.

        Arguments passed to this function override defaults for the algorithm.
        '''
        assert self._default_hyperparams is not None, (
            'BUG: Algorithm._default_hyperparams should be resolved in __init__.')
        default_hyperparams = self._default_hyperparams.copy()
        default_hyperparams.update(**overrides)
        return default_hyperparams

    def _param_from_json(self, hyperparam_name: str, json_value: Any) -> Any:
        '''Parse the value for the given hyperparameter from the problem definition JSON.

        The default parser simply returns the value we parsed from JSON. Not all
        hyperparameter values are strings, numbers, lists, or dictionaries. Some
        are full python objects.
        '''
        # We really want to change the text of the KeyError.
        # pylint: disable=raise-missing-from
        if hyperparam_name in self._hyperparam_lookup:
            try:
                return self._hyperparam_lookup[hyperparam_name][json_value]
            except KeyError:
                raise KeyError(f'for {hyperparam_name}, could not find a parsed python value for'
                               f' json value {json_value} in hyperparam lookup table.  Only'
                               f' found {self._hyperparam_lookup[hyperparam_name].keys()}.')
        return json_value

    def param_to_json(self, hyperparam_name: str, python_value: Any) -> Any:
        '''Represent a hyperparam as a JSON serializabe structure.'''
        if hyperparam_name in self._hyperparam_lookup:
            for json_value, table_python_value in self._hyperparam_lookup[hyperparam_name].items():
                if python_value == table_python_value:
                    return json_value
            raise KeyError(f'for {hyperparam_name}, could not find a json value for '
                           f'python value {python_value} in hyperparam lookup table. '
                           f'Only found {self._hyperparam_lookup[hyperparam_name].values()}.')
        return python_value

    def bind(self, hyperparam_name: str, param_range: ParamRange) -> Dict[str, Any]:
        '''Bind names to final values for params.

        Returns: {'valuestr': value}
        '''
        return BinderFactory().build(
            param_range.method, parse_value=self._param_from_json).bind(
                hyperparam_name=hyperparam_name, prange=param_range.range)


class AlgorithmStub(Algorithm):
    '''Stub algorithm'''

    def instantiate(self, **hyperparams) -> Any:
        return None


class AlgorithmCatalog(Catalog[Algorithm], metaclass=abc.ABCMeta):
    '''Base class for algorithm catalogs'''
    # pylint: disable=unused-argument
    # These unused arguments are to make signatures match. quack.
    def __init__(self, communicator_catalog: Optional[CommunicatorCatalog] = None,
                 discoverer_catalog: Optional[DiscovererCatalog] = None, **kwargs):
        # This makes mypy happy. We make the signature compatable with MemoryAlgorithmCatalog.
        super().__init__(**kwargs)


class MemoryAlgorithmCatalog(MemoryCatalog[Algorithm], AlgorithmCatalog):
    '''An Algorithm catalog that has load and register.'''


class AlgorithmCatalogStub(MemoryCatalog[Algorithm], AlgorithmCatalog):
    '''stub'''



'''This is the executable pipeline for the "simple" executor.'''
from typing import Iterable, List, Optional

from ...algorithms.impl.algorithm_instance import AlgorithmInstance
from ...algorithms.impl.fittable_algorithm_instance import UntrainedError
from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import Designator
from ...instantiator.executable_pipeline import (ExecutablePipeline,
                                                 PipelineResult,
                                                 FitError)
from ...templates.impl.parallel_step import ParallelStep
from ...wrangler.dataset import Dataset, DatasetKeys
from ..cucumber import JarOfCucumbers
from ..executor_kind import ExecutorKind
from .simple_executable_step import SimpleExecutableStep
from .simple_parallel_executable_step import SimpleParallelExecutableStep

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class SimpleExecutablePipeline(ExecutablePipeline):
    '''A pipeline for the 'simple' executor.'''
    _kind = ExecutorKind('simple')
    _steps: List[SimpleExecutableStep]

    def __init__(self,
                 pipeline: BoundPipeline):
        trained = False
        self._pipeline = pipeline
        self._steps = []
        for step in pipeline.steps:
            if isinstance(step, ParallelStep):
                executable_step: SimpleExecutableStep = (
                    SimpleParallelExecutableStep(step, SimpleExecutablePipeline))
            else:
                executable_step = SimpleExecutableStep(step)
                if executable_step.trained:
                    trained = True
            self._steps.append(executable_step)
        if trained:
            # Mainly, this catches the subpipelines of parallel steps
            # which are not otherwise marked as trained.
            self.set_trained()

    def set_trained(self):
        super().set_trained()
        for step in self._steps:
            step.set_trained()

    @property
    def locked(self) -> bool:
        '''Does this pipeline contain a model that is currently locked?'''
        return any(step.locked for step in self._steps)

    def converged(self) -> bool:
        '''Is every distributed model in this pipeline converged?

        Always returns True for pipelines with no distributed models.
        '''
        return all(step.converged() for step in self._steps)

    def set_fit_error(self, err: FitError) -> None:
        '''Set this pipeline as throwing an error at fit time'''
        self._fit_error = err

    def fit(self, dataset: Optional[Dataset]) -> JarOfCucumbers:
        '''Fit models to dataset.

        Args:
            dataset: The training dataset.

        Returns:
            A JarOfCucumbers
            (mapping from step designators to cucumbers for each step).
        '''
        # Pipelines set themselves as trained first, so that they are
        #   marked trained even if they throw an exception along the way
        self._trained = True

        # If there is a fit error recorded, remove it.
        self._fit_error = None

        # If dataset is None, just call fit(None) on all the steps.
        # This allows distributed pipelines to fit themselves with
        # neighbor data.
        result = None
        inp = dataset
        for step in self._steps:
            step.fit(dataset=inp)
            if inp is not None:
                result = step.predict(dataset=inp)
            inp = result
        if inp is not None:
            return self.cucumberize_all()
        return JarOfCucumbers()

    def predict(self, dataset: Optional[Dataset]) -> PipelineResult:
        '''Run prediction for a complete pipeline.'''
        if not self._trained:
            raise UntrainedError(f'pipeline "{self.name}" needs to be fit before it can predict')
        if dataset is None:
            inp = None
        else:
            inp = dataset.output()
        if self._fit_error is not None:
            if inp is None:
                inp = Dataset(metadata=None, **{DatasetKeys.ERROR.value: self._fit_error})
            else:
                inp[DatasetKeys.ERROR.value] = self._fit_error
        else:
            if inp is not None and dataset is not None:
                inp.update(dataset)
            else:
                inp = dataset
            for step in self._steps:
                result = step.predict(dataset=inp)
                inp = result
        return PipelineResult(executable_pipeline=self, prediction=inp)

    def cucumberize_all(self, parent_designator: Optional[Designator] = None) -> JarOfCucumbers:
        '''Cucumberize all steps that hold algorithms, and return the result.

        This has the same output as fit() but requires the pipeline to be fit already.

        The parent_designator argument is not None iff this pipeline is a subpipeline inside
            a parallel step.  It is there so that steps in subpipelines can be saved with
            their parent's designator.
        '''
        if not self._trained:
            raise UntrainedError(
                f'pipeline "{self.name}" needs to be fit before it can cucumberize_all')

        models = JarOfCucumbers()
        for step in self._steps:
            models.update(
                step.cucumberize_all(
                    parent_designator or self.designator))
        return models

    def start(self) -> None:
        '''Start the pipeline.'''
        for step in self._steps:
            step.start()

    def stop(self) -> None:
        '''Stop the pipeline.'''
        for step in self._steps:
            step.stop()

    @property
    def all_instances(self) -> Iterable[AlgorithmInstance]:
        '''Iterate through all algorithm instances.'''
        for step in self._steps:
            yield from step.all_instances



'''Decider that requires one of its subdeciders to decide True.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Mapping
from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DecidersConfig
from ..neighbor_manager.node_id import NodeID

from .impl.decider_catalog import DeciderCatalog

from .boolean import BooleanDeciderConfig, BooleanDecider


class OrDeciderConfig(BooleanDeciderConfig):
    '''Configuration for OrDecider.'''

    def __str__(self) -> str:
        return f'or({[str(c) for c in self._clause]}'


class OrDecider(BooleanDecider):
    '''Decider that decides True if any of its subdeciders decides True.'''
    name = 'or'
    tags = {}
    _config: OrDeciderConfig

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        return any(self._all_decisions(
            my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors))


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(OrDecider)
    DecidersConfig.register(name=OrDecider.name, config_type=OrDeciderConfig)



'''Sends and recieves messages using UDP sockets.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
from queue import Queue
import socket
from typing import Callable, Dict, List, Optional, Set, Tuple

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..config_components.distributed_config import CommunicatorConfig, DistributedConfig
from ..config_components.impl.config_component import ConfigComponent
from ..neighbor_manager.event import NewNeighbor, Recv
from ..neighbor_manager.node_id import NodeID
from ..wrangler.exception_thread import ExceptionThread
from ..wrangler.logger import Level, Logger

from .impl.endpoint import Endpoint
from .impl.ip_endpoint import IPEndpoint
from .impl.communicator import Communicator
from .impl.communicator_catalog import CommunicatorCatalog

logger = Logger(__file__, level=Level.DEBUG).logger()


class SocketsConfig(ConfigComponent):
    '''Configuration for the SocketsCommunicator.'''
    name: str = 'sockets'
    tags: Dict[str, List[str]] = {}

    class Constants(Enum):
        '''Keys below the top level.'''

    class Keys(AEnum):
        '''Valid keys for the top level.'''
        NODES_AND_ENDPOINTS = 'nodes_and_endpoints'
        LISTENER_TIMEOUT = 'listener_timeout'

    class Defaults(Enum):
        '''Default values'''
        LISTENER_TIMEOUT = 2.0

    def required_keys(self) -> Set[str]:
        return set()

    @property
    def nodes_and_endpoints(self) -> List[Tuple[NodeID, Endpoint]]:
        '''Get a preconfigured relationship between nodes and endpoints.'''
        retval: List[Tuple[NodeID, Endpoint]] = []
        if not self._exists(
            self.Keys.NODES_AND_ENDPOINTS
        ):
            return retval
        for node, endpoint in self._get(
            self.Keys.NODES_AND_ENDPOINTS
        ):
            node_id = NodeID(int(node))
            # We only know how to build IP endpoints at the moment.
            ip, port = endpoint
            retval.append(
                (node_id, IPEndpoint(ip, int(port)))
            )
        return retval

    @property
    def listener_timeout(self) -> float:
        '''Get the listener timeout.'''
        return float(self._get_with_default(
            self.Keys.LISTENER_TIMEOUT,
            dflt=self.Defaults.LISTENER_TIMEOUT.value
        ))


def _socket_listener(queue: Queue,
                     sock: socket.socket,
                     assign_node_id: Callable[[IPEndpoint], NodeID],
                     stop: Callable[[], bool],
                     timeout: int,
                     known_neighbors: Set[NodeID]):

    while True:
        try:
            if stop():
                sock.close()
                return
            payload, addr = sock.recvfrom(65536)
            endpoint = IPEndpoint(host=addr[0], port=addr[1])
            neighbor = assign_node_id(endpoint)
            if neighbor not in known_neighbors:
                known_neighbors.add(neighbor)
                queue.put(NewNeighbor(neighbor=neighbor), timeout=timeout)
            queue.put(Recv(neighbor, payload), timeout=timeout)
        except socket.timeout:
            pass


class SocketsCommunicator(Communicator):
    '''Sends and recieves messages using UDP sockets.'''
    name: str = 'sockets'
    tags: Dict[str, List[str]] = {}
    _stop: bool
    _node_to_endpoint: Dict[NodeID, IPEndpoint]
    _endpoint_to_node: Dict[IPEndpoint, NodeID]
    _sock: Optional[socket.socket]
    _next_node_id: NodeID
    _listener: Optional[ExceptionThread] = None

    def __init__(self,
                 known_neighbors: Set[NodeID],
                 my_id: NodeID,
                 distributed: DistributedConfig):
        self._node_to_endpoint = {}
        self._endpoint_to_node = {}
        self._stop = False
        self._sock = None
        self._next_node_id = NodeID(1)
        for (node, endpoint) in distributed.communicator.sockets.nodes_and_endpoints:  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
            assert isinstance(endpoint, IPEndpoint), (
                f'BUG: endpoint={endpoint} is if type={type(endpoint)}, not IPEndpoint'
            )
            self._node_to_endpoint[node] = endpoint
            self._endpoint_to_node[endpoint] = node
            if node >= self._next_node_id:
                self._next_node_id = NodeID(node + 1)

        super().__init__(my_id=my_id, known_neighbors=known_neighbors, distributed=distributed)

    def stop(self):
        '''Tell threads to terminate themselves.'''
        if self._listener is not None:
            logger.debug('Stopping SocketsCommunicator: %s', self._listener)
            self._stop = True
            self._listener.join()
            self._listener = None

    def assign_node_id(self, endpoint: IPEndpoint) -> NodeID:
        '''Assign a node ID to the given endpoint (or find one that was already assigned).'''
        if endpoint not in self._endpoint_to_node:
            node = self._next_node_id
            self._next_node_id = NodeID(self._next_node_id + 1)
            self._endpoint_to_node[endpoint] = node
            self._node_to_endpoint[node] = endpoint
        if self._endpoint_to_node[endpoint] not in self._known_neighbors:
            logger.debug(
                'Node %s on port %s discovered neighbor node %s on port %s',
                self._my_id, self._node_to_endpoint[self._my_id],
                self._endpoint_to_node[endpoint], endpoint)
        return self._endpoint_to_node[endpoint]

    def lookup_endpoint(self, node: NodeID) -> IPEndpoint:
        '''Find the IPEndpoint that corresponds to the node.'''
        return self._node_to_endpoint[node]

    def start(self,
              queue: Queue,
              timeout: Optional[float] = None) -> None:
        '''Start the communicator.'''
        if timeout is None:
            timeout = self._distributed.communicator.sockets.listener_timeout  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        self._sock = socket.socket(family=socket.AF_INET, type=socket.SOCK_DGRAM)
        self._sock.settimeout(timeout)
        my_endpoint = self._node_to_endpoint[self._my_id]
        logger.debug('UDP SocketsCommunicator binding to %s:%s', my_endpoint.host, my_endpoint.port)
        try:
            self._sock.bind((my_endpoint.host, my_endpoint.port))
        except OSError as err:
            raise OSError(f'Error binding to {my_endpoint.host}:{my_endpoint.port}') from err
        self._stop = False
        self._listener = ExceptionThread(target=_socket_listener, kwargs={
            'queue': queue,
            'sock': self._sock,
            'assign_node_id': self.assign_node_id,
            'stop': lambda: self._stop,
            'timeout': timeout,
            'known_neighbors': self._known_neighbors,
        })
        self._listener.start()

    def send(self,
             dest_ids: List[NodeID],
             payload: bytes) -> int:
        '''Send a message to the node with dest_id.

        Return value is the size of the message sent.
        '''
        assert self._sock is not None, (
            'BUG: send() called before start().'
        )
        retval = 0
        for dest_id in dest_ids:
            dest_endpoint = self._node_to_endpoint[dest_id]
            logger.debug('UDP sockets sending to dest_id=%s, dest_endpoint=%s',
                         dest_id, dest_endpoint)
            try:
                retval += self._sock.sendto(payload, dest_endpoint.as_tuple)
                logger.log(Level.VERBOSE, 'payload size: %s', len(payload))
            # PermissionError can occur during shutdown. Just absorb it.
            except PermissionError as err:
                logger.error('%s: permission error sending payload: %s', err, len(payload))
            except OSError as err:
                logger.error('payload size too large to send: %s', len(payload))
                raise err

        return retval

    def send_all(self, payload: bytes) -> int:
        '''Send a message to all known neighbors.

        Return value is a count of neighbors we sent to.
        '''
        dest_ids = list(self._known_neighbors)
        logger.debug('UDP Socket sending to dest_ids = %s', dest_ids)
        self.send(dest_ids=dest_ids, payload=payload)
        return len(dest_ids)


def register(catalog: CommunicatorCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(SocketsCommunicator)
    CommunicatorConfig.register(name="sockets", config_type=SocketsConfig)



'''The PipelineTemplate object.

This holds the steps in an AutonML pipeline template, and any
supporting metadata.
'''
from typing import Any, Dict, List, Optional, Union, Iterable

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...algorithms.impl.algorithm import Algorithm, AlgorithmCatalog, AlgorithmCatalogStub
from ...catalog.catalog_element_mixin import CatalogElementMixin
from ...generator.designator import Designator
from .parallel_step import ParallelStep
from .pipeline_step import PipelineStep, GeneratorInterface
from .query_step import QueryStep


class Error(Exception):
    '''Base exception for PipelineTemplate objects.'''


class UndefinedError(Error):
    '''An undefined variable was referenced.'''


class PipelineTemplate(CatalogElementMixin):
    '''A template for a group of pipelines.

    The :doc:`/_autosummary/ngautonml.generator.generator.Generator`
    converts a template into several
    :doc:`/_autosummary/ngautonml.generator.bound_pipeline.BoundPipeline`\\ s.

    The ``Generator`` creates a new ``BoundPipeline`` for each combination of
    algorithms returned by all the query steps in the template.

    The ``Searcher`` creates new ``BoundPipeline``s out of those created
    by the ``Generator``, using hyperparameter search.

    Finally, the ``Instantiator`` turns all of these ``BoundPipelines``
    into ``ExecutablePipelines`` to be run by the ``Executor``.

    :name: the unique primary key for this template in the template catalog.

    :tags: secondary keys for this template in the template catalog.
        If specified, they override all the tags in the declaration of the class.

    '''
    _name: str = 'unnamed'
    _steps: List[PipelineStep]
    _algorithm_catalog: AlgorithmCatalog
    _generator: Optional[GeneratorInterface]

    def __init__(self,    # pylint: disable=too-many-arguments
                 name: Optional[str] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 algorithm_catalog: Optional[AlgorithmCatalog] = None,
                 generator: Optional[GeneratorInterface] = None):
        super().__init__(name=name, tags=tags)
        self._steps = []
        self._algorithm_catalog = algorithm_catalog or AlgorithmCatalogStub()
        self._generator = generator

    @property
    def steps(self) -> List[PipelineStep]:
        '''Returns the steps of this pipeline.'''
        return self._steps[:]

    def step(self,
             model: Optional[Algorithm] = None,
             serialized_model: Optional[bytes] = None,
             **overrides) -> PipelineStep:
        '''Add a step to the end of the template.'''
        step = PipelineStep(
            model=model,
            serialized_model=serialized_model,
            algorithm_catalog=self._algorithm_catalog,
            generator=self._generator,
            **overrides)
        self._steps.append(step)
        return step

    def new(self, name: str) -> 'PipelineTemplate':
        '''Create a new descendant PipelineTemplate

        This is for building parallel threads.
        '''
        child = PipelineTemplate(name=f'{name}',
                                 algorithm_catalog=self._algorithm_catalog,
                                 generator=self._generator)
        return child

    def parallel(self, **pipelines: 'PipelineTemplate') -> PipelineStep:
        '''Add a parallel step, which contains 2 or more subpipelines.

        Parallel steps allow pipelines to do different things with different parts of the data.
        For example, running preprocessors on the covariates but not the target column.

        * ``**pipelines``: A dictionary of ``{output_key: pipeline}``.
          The ``output_key`` for a subpipeline is where the parallel step stores
          the results in the Dataset,
          and can be any string, naming the subpipeline.
          The pipelines should be generated from ``PipelineTemplate().new()``.

        **Returns**: The new parallel step.
        '''
        retval = ParallelStep(model=None,
                              algorithm_catalog=self._algorithm_catalog,
                              generator=self._generator,
                              **pipelines)
        self._steps.append(retval)
        return retval

    def query(self, **tags: Union[str, Iterable[str]]) -> PipelineStep:
        '''Creates a query pipeline step and adds it to the template.

        A query step queries the ``AlgorithmCatalog`` by tag,
        and returns a set of ``Algorithms``.
        The ``Generator`` creates a new ``BoundPipeline``
        for each element of the cartesian product of
        all the query steps in the template.

        * ``**tags``: looks like ``tag_type = tag_value`` or
          ``tag_type = (tag_value1, tag_value2)``

        Queries for algorithms that match at least one specified tag value
        for all specified tag types.
        '''
        retval = QueryStep(algorithm_catalog=self._algorithm_catalog, **tags)
        self._steps.append(retval)
        return retval

    def generate(self) -> Dict[Designator, Any]:
        '''Generate all the bound pipelines for this pipeline template.'''
        return {}



'''Generates a constructor for an SklearnLoader.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=duplicate-code,too-many-arguments
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Type
import importlib
import random

from numpy.random import RandomState  # pylint: disable=no-name-in-module  # This just isn't true.
import pandas as pd

from ...config_components.impl.config_component import InvalidValueError
from ...wrangler.constants import Defaults
from ...wrangler.dataset import Dataset
from ..impl.data_loader_catalog import DataLoaderCatalog
from ..impl.dataframe_loader import DataframeLoader
from .sklearn_loader_interface import SklearnLoaderInterface


def _load_train_impl(self: SklearnLoaderInterface) -> Dataset:
    '''Implementation for _load_train() in the SklearnLoader we generate'''
    # pylint: disable=protected-access
    params = deepcopy(self._config.params)
    if self.tags.get('supports_random_seed', ['false'])[0] == 'true':
        if 'random_seed' in params:
            params['random_state'] = RandomState(params.pop('random_seed'))
        elif params.get('random_state', None) is None:
            params['random_state'] = RandomState(Defaults.SEED)

    if self.tags.get('uses_return_X_y', ['false'])[0] == 'true':
        params['return_X_y'] = True

    if self.tags.get('uses_as_frame', ['false'])[0] == 'true':
        params['as_frame'] = True

    restrict_to_n_features = -1
    if 'restrict_features' in params:
        restrict_to_n_features = params.pop('restrict_features')

    test_size = -1
    if 'test_size' in params:
        test_size = params.pop('test_size')

    (x, y) = self._impl._loader(**params)
    x = pd.DataFrame(x)
    # Make sure all column names are strings
    #   (casting to Index to make mypy happy)
    x.columns = pd.core.indexes.base.Index([str(n) for n in x.columns])

    if restrict_to_n_features >= 0:
        restrict_to_n_features = min(restrict_to_n_features, x.shape[1])
        x = x.iloc[:, :restrict_to_n_features]

    if test_size > x.shape[0]:
        raise InvalidValueError(
            f'test_size is set to {test_size}, but dataset {self.name} '
            f'only has {x.shape[0]} observations.'
        )

    target = self._config.metadata.target
    target_name = target.name if target is not None else 'y'

    if test_size > 0:
        x_test = x.iloc[-test_size:, :]
        y_test = y[-test_size:]
        test_data = x_test
        test_data[target_name] = y_test
        self._test_data = test_data

        x = x.iloc[:-test_size, :]
        y = y[:-test_size]

    train_df = x
    train_df[target_name] = y

    retval = self._dataset(data=train_df)
    return retval


def _load_test_impl(self: SklearnLoaderInterface) -> Optional[Dataset]:
    '''Implementation for _load_test() in the SklearnLoader we generate'''
    # pylint: disable=protected-access
    if self._test_data is not None:
        return self._dataset(data=self._test_data)
    return None


class SklearnLoaderGenerator():
    '''Generates a constructor for an SklearnLoader.'''
    _loader: Callable[..., Any]

    def __init__(self,
                 name: str,
                 loader: Optional[type] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 **hyperparams: Any):
        self.name = name
        if loader is None:
            loader = self._load_module(name)
        self._loader = loader
        if tags is not None:
            self.tags = tags.copy()
        else:
            self.tags = {}
        super().__init__(**hyperparams)

    def _random_name(self) -> str:
        # We don't actually care about collisions as we never look
        # these constructors up by name.
        return f'GeneratedSklearnLoader{random.randint(0, 999999999)}'

    def _load_module(self, name: str):
        # Split name into module part (e.g. sklearn.linear_model)
        # and constructor part (e.g. LinearRegression)
        parts = name.split('.')
        constructor_part = parts[-1]
        module = importlib.import_module('.'.join(parts[:-1]))
        # Load the constructor.
        return getattr(module, constructor_part)

    def new_class(self) -> Type[DataframeLoader]:
        '''Generate a constructor for an SklearnLoader.'''
        return type(self._random_name(), (DataframeLoader, ), {
            "name": self.name,
            "tags": self.tags.copy(),
            "_load_train": _load_train_impl,
            "_load_test": _load_test_impl,
            "_test_data": None,
            "_impl": self,
        })


def register(catalog: DataLoaderCatalog):  # pylint: disable=unused-argument
    '''Nothing to register.

    All subclasses of SklearnAlgorithm are registered in sklearn_algorithms.py
    '''



'''These are the kinds of executable pipelines for different executors.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, Iterable, Optional

import pandas as pd


from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..executor.cucumber import JarOfCucumbers
from ..executor.executor_kind import ExecutorKind
from ..generator.bound_pipeline import BoundPipeline, BoundPipelineStub
from ..generator.designator import Designator
from ..splitters.impl.splitter import SplitDataset
from ..wrangler.constants import OutputColName
from ..wrangler.dataset import Dataset, DatasetKeys


class PipelineExecutionError(Exception):
    '''Base class for errors thrown by the executor'''
    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return self.__class__.__name__ + '(' + ',\n'.join(self.args) + ')'


class FitError(PipelineExecutionError):
    '''A pipeline threw an error during fit.'''


class PredictError(PipelineExecutionError):
    '''A pipeline threw an error during predict'''


class ExecutablePipeline(metaclass=abc.ABCMeta):
    '''Base class for all executable pipelines.

    An ``ExecutablePipeline`` can be executed by a specific kind of
    ``Executor``, and is created from a ``BoundPipeline``
    by the corresponding ``Instantiator``.

    All algorithms and hyperparameters are bound to single values
    (so no query steps or hyperparam search spaces).
    '''
    _kind: ExecutorKind
    _pipeline: BoundPipeline
    _trained: bool = False
    _fit_error: Optional[FitError] = None

    @property
    def kind(self):
        '''What kind of executor are we built for?

        e.g. json, python_script, jupyter_notebook, AirFlow Class, stub
        '''
        return self._kind

    @property
    def name(self) -> str:
        '''The name of the underlying pipeline'''
        return self._pipeline.name

    @property
    def designator(self) -> Designator:
        '''The designator for this instance of an executable pipeline

        Always the same as that of its bound pipeline
        '''
        return self.bound.designator

    @property
    def bound(self) -> BoundPipeline:
        '''The bound pipeline this was compiled from'''
        return self._pipeline

    @property
    def locked(self) -> bool:
        '''Does this pipeline contain a model that is currently locked?'''
        raise NotImplementedError(
            'There is currently no way for a base ExecutablePipeline to know about its models.  '
            'Perhaps this should be a SimpleExecutablePipeline?')

    @property
    def trained(self) -> bool:
        '''Is this pipeline trained?'''
        return self._trained

    def set_trained(self):
        '''Force the pipeline to trained.'''
        self._trained = True

    def converged(self) -> bool:
        '''Is every distributed model in this pipeline converged?

        Always returns True for pipelines with no distributed models.
        '''
        return True

    @abc.abstractmethod
    def fit(self, dataset: Dataset) -> JarOfCucumbers:
        '''Fit models to dataset.

        Args:
            dataset: The training dataset.

        Returns:
            A JarOfCucumbers
            (mapping from step designators to cucumbers for each step).
        '''

    @abc.abstractmethod
    def predict(self, dataset: Dataset) -> 'PipelineResult':
        '''Get predictions for this pipeline on this dataset.'''

    @abc.abstractmethod
    def cucumberize_all(self) -> JarOfCucumbers:
        '''Cucumberize all steps that hold algorithms, and return the result.

        This has the same output as fit() but requires the pipeline to be fit already.
        '''

    @abc.abstractmethod
    def start(self) -> None:
        '''Start all pipeline threads.'''

    @abc.abstractmethod
    def stop(self) -> None:
        '''Stop all pipeline threads.'''

    @property
    @abc.abstractmethod
    def all_instances(self) -> Iterable[AlgorithmInstance]:
        '''Iterate through all algorithm instances.'''

    def __str__(self):
        return f'{{{self._pipeline.designator}, executor: {self.kind}}}'


class ExecutablePipelineStub(ExecutablePipeline):
    '''stub'''
    _cucumbers: JarOfCucumbers

    def __init__(self,
                 kind: Optional[ExecutorKind] = None,
                 pipeline: Optional[BoundPipeline] = None,
                 trained: bool = False,
                 cucumbers: Optional[JarOfCucumbers] = None):
        self._kind = kind or ExecutorKind('stub_executor_kind')
        self._pipeline = pipeline or BoundPipelineStub(name='stub_pipeline')
        self._trained = trained
        self._cucumbers = cucumbers or JarOfCucumbers({})

    def fit(self, dataset: Dataset) -> JarOfCucumbers:
        return JarOfCucumbers()

    def predict(self, dataset: Dataset) -> 'PipelineResult':
        return PipelineResult(prediction=dataset, executable_pipeline=self)

    def cucumberize_all(self) -> JarOfCucumbers:
        return self._cucumbers

    def stop(self) -> None:
        pass

    def start(self) -> None:
        pass

    @property
    def all_instances(self) -> Iterable[AlgorithmInstance]:
        return []

    @property
    def locked(self) -> bool:
        return False

    def converged(self) -> bool:
        '''Has the model converged?

        Non-distributed models are always converged.
        '''
        return True


class PipelineResultError(Exception):
    '''Base class for all errors associated with this file.'''


# TODO(Merritt): put this in own file
# (may need a separate InternalPipelineResult and ExternalPipelineResult)
class PipelineResult(metaclass=abc.ABCMeta):
    '''Base class for objects that represent the result of a pipeline run.

    At a minimum contains a pipeline that was run and a dataset resulting from that run.
    Subclasses may contain additional information.
    '''
    _prediction: Optional[Dataset]
    _bound_pipeline: Optional[BoundPipeline] = None
    _executable_pipeline: Optional[ExecutablePipeline] = None
    _split_dataset: Optional[SplitDataset] = None

    def __init__(self,
                 prediction: Optional[Dataset],
                 bound_pipeline: Optional[BoundPipeline] = None,
                 executable_pipeline: Optional[ExecutablePipeline] = None,
                 split_dataset: Optional[SplitDataset] = None):

        if bound_pipeline is None:
            assert executable_pipeline is not None, (
                'BUG: both bound and executable pipelines are None'
                ' in PipelineResult instantiator.')

        self._prediction = prediction
        self._bound_pipeline = bound_pipeline
        self._executable_pipeline = executable_pipeline
        self._split_dataset = split_dataset
        if (self._split_dataset is not None
            and self._split_dataset.ground_truth is not None
                and DatasetKeys.PREDICTIONS.value in (self._prediction or {})):
            self._add_ground_truth()

    def _add_ground_truth(self) -> None:
        '''Display ground truth next to train predictions.'''
        assert self._split_dataset is not None
        assert self._split_dataset.ground_truth is not None
        assert self.prediction is not None, (
            'BUG: attempt to add ground truth to a pipeline result with no prediction of None.'
        )
        assert isinstance(self.prediction[DatasetKeys.PREDICTIONS.value], pd.DataFrame), (
            'BUG: can only add ground truth to predictions if predictions is a dataframe. '
            f'instead found f{self.prediction[DatasetKeys.PREDICTIONS.value]}')

        if OutputColName.GROUND_TRUTH.value in self.prediction.predictions.columns:
            # Predictions dataframe already contains ground truth
            return
        gt_df = self._split_dataset.ground_truth[DatasetKeys.GROUND_TRUTH.value]
        assert isinstance(gt_df, pd.DataFrame), (
            'BUG: can only add ground truth to predictions if ground truth is a dataframe. '
            f'instead found {gt_df} of type {type(gt_df)}')
        target = self._split_dataset.ground_truth.metadata.target
        assert target is not None, (
            'BUG: attempt to add ground truth with no target.')
        new_df = self.prediction.predictions.copy(deep=True)
        # Align the ground truth with the prediction dataframe.
        new_df.reset_index(inplace=True, drop=True)
        gt_df.reset_index(inplace=True, drop=True)
        new_df[OutputColName.GROUND_TRUTH.value] = gt_df[target.name]
        self.prediction.predictions = new_df

    @property
    def prediction(self) -> Optional[Dataset]:
        '''Dataset resulting from the pipeline run'''
        return self._prediction

    @property
    def family_designator(self) -> Designator:
        '''Family designator of BoundPipeline associated with this pipeline run'''
        return self.bound_pipeline.family_designator

    @property
    def bound_pipeline(self) -> BoundPipeline:
        '''Bound pipeline associated with this pipeline run.'''
        if self._executable_pipeline is not None:
            return self._executable_pipeline.bound
        assert self._bound_pipeline is not None, (
            'BUG: both bound and executable pipelines are None'
            ' in PipelineResult bound_pipeline property.')
        return self._bound_pipeline

    @property
    def executable_pipeline(self) -> Optional[ExecutablePipeline]:
        '''Trained executable pipeline associated with this pipeline run, if applicable.'''
        return self._executable_pipeline

    @property
    def split_dataset(self) -> Optional[SplitDataset]:
        '''Split dataset that was passed to the CrossValidator, if applicable'''
        return self._split_dataset

    def __str__(self) -> str:
        return f'pipeline: {self.bound_pipeline!s}\nprediction:\n{self.prediction!s}'

    def __repr__(self) -> str:
        return self.__str__()


class PipelineResults(Dict[Designator, PipelineResult]):
    '''Multiple PipelineResults identified by Designators.'''

    def _some_result(self) -> PipelineResult:
        '''Choose an arbitrary result and return it.'''
        return list(self.values())[0]

    def infer_ground_truth(self) -> Optional[Dataset]:
        '''Attempts to infer ground truth using an arbitrary PipelineResult.'''
        split_dataset = self._some_result().split_dataset
        if split_dataset is not None:
            return split_dataset.ground_truth
        return None

    @property
    def predictions(self) -> Dict[Designator, Optional[Dataset]]:
        '''Datasets resulting from the pipeline runs'''
        return {k: v.prediction for k, v in self.items()}

    @property
    def bound_pipelines(self) -> Dict[Designator, BoundPipeline]:
        '''Bound Pipelines associated with the pipeline runs'''
        return {k: v.bound_pipeline for k, v in self.items()}

    @property
    def executable_pipelines(self) -> Dict[Designator, ExecutablePipeline]:
        '''Trained executable pipelines associated with the pipeline runs.

        Throws an error if not all runs have executable pipelines
        '''
        retval: Dict[Designator, ExecutablePipeline] = {}
        for des, pipe in self.items():
            pipeline = pipe.executable_pipeline
            if pipeline is None:
                raise PipelineResultError(
                    'Attempt to extract executable pipeline '
                    'from a PipelineResult that does not have one'
                    f' (designator: {pipe.bound_pipeline.designator}).')
            retval[des] = pipeline
        return retval

    def __str__(self) -> str:
        retval = 'Pipeline Results:\n'
        for pipe in self.values():
            retval = f'{retval}{pipe}\n'
        return retval



'''Tests for abstractor.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=protected-access,too-many-statements,duplicate-code,too-many-locals

from copy import deepcopy
import getpass
import os
from pathlib import Path

import pandas as pd

from ....config_components.distributed_config import DistributedConfig
from ....conftest import Clearer, Waiter
from ....problem_def.problem_def import ProblemDefinition
from ....wrangler.base_port import BasePort
from ....wrangler.dataset import Dataset
from ....wrangler.logger import Logger
from ....wrangler.wrangler import Wrangler

from ...distributed.auton_decision_tree import AutonDecisionTreeModel
from ...impl.synchronous import advance

from ..share import ShareAbstractor


logger = Logger(__file__, to_stdout=False).logger()


base_port = BasePort()


LOOPBACK = '127.0.0.1'
SUNNY_DAY_SENDER = base_port.next()
SUNNY_DAY_RECEIVER = base_port.next()


def test_sunny_day(wait_til_all_fit: Waiter,
                   assert_no_exceptions: Clearer) -> None:
    '''One node shares its data to its neighbor.'''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': [2], '2': [1]}
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, SUNNY_DAY_SENDER)),
                    ('2', (LOOPBACK, SUNNY_DAY_RECEIVER))
                ]
            }
        },
        'my_id': 1,
    }

    dut_sender = ShareAbstractor().instantiate(
        distributed=DistributedConfig(clause=distributed_clause))

    sender_dataset = Dataset()
    sender_dataset.dataframe = pd.DataFrame({'a': [1], 'b': [2], 'c': [5]})

    receiver_clause = deepcopy(distributed_clause)
    receiver_clause['my_id'] = 2
    dut_receiver = ShareAbstractor().instantiate(
        distributed=DistributedConfig(clause=receiver_clause))

    try:
        dut_sender.start()
        dut_receiver.start()
        dut_sender.fit(sender_dataset)

        # Wait for the message to arrive and get processed.
        wait_til_all_fit([dut_sender, dut_receiver])

        assert_no_exceptions([dut_sender, dut_receiver])

        result = dut_receiver.synthesize()

        assert result.dataframe.equals(sender_dataset.dataframe)

    finally:
        dut_sender.stop()
        dut_receiver.stop()


BIDIRECTIONAL_SENDER = base_port.next()
BIDIRECTIONAL_RECEIVER = base_port.next()


def test_bidirectional(assert_no_exceptions: Clearer) -> None:
    '''One node shares its data to its neighbor.'''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': [2], '2': [1]}
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, BIDIRECTIONAL_SENDER)),
                    ('2', (LOOPBACK, BIDIRECTIONAL_RECEIVER))
                ]
            }
        },
        'my_id': 1,
    }

    dut_sender = ShareAbstractor().instantiate(
        distributed=DistributedConfig(clause=distributed_clause), synchronous=True)

    sender_dataset = Dataset()
    sender_dataset.dataframe = pd.DataFrame({'a': [1], 'b': [2]})

    receiver_clause = deepcopy(distributed_clause)
    receiver_clause['my_id'] = 2
    dut_receiver = ShareAbstractor().instantiate(
        distributed=DistributedConfig(clause=receiver_clause), synchronous=True)

    receiver_dataset = Dataset()
    receiver_dataset.dataframe = pd.DataFrame({'a': [3], 'b': [4]})

    try:
        dut_sender.start()
        dut_receiver.start()
        dut_sender.fit(sender_dataset)
        dut_receiver.fit(receiver_dataset)

        # Wait for the messages to arrive and get processed.
        for _ in range(10):
            advance([dut_sender, dut_receiver])

        assert_no_exceptions([dut_sender, dut_receiver])

        result_sender = dut_sender.synthesize()
        result_sender_df = result_sender.dataframe
        result_sender_df = result_sender_df.sort_values(
            by=result_sender_df.columns.tolist(), ignore_index=True)

        result_receiver = dut_receiver.synthesize()
        result_receiver_df = result_receiver.dataframe
        result_receiver_df = result_receiver_df.sort_values(
            by=result_receiver_df.columns.tolist(), ignore_index=True)

        pd.testing.assert_frame_equal(result_receiver_df, result_sender_df)

        want_df = pd.DataFrame({'a': [1, 3], 'b': [2, 4]})
        want_df = want_df.sort_values(
            by=want_df.columns.tolist(), ignore_index=True)

        pd.testing.assert_frame_equal(want_df, result_receiver_df)
        pd.testing.assert_frame_equal(want_df, result_sender_df)

    finally:
        dut_sender.stop()
        dut_receiver.stop()


BOOTSTRAP_SERVERS = os.environ.get('BOOTSTRAP_SERVERS', 'localhost:9092')


def topic_name(topic: str) -> str:
    '''Generate a unique topic name for the given test.'''
    return f'{Path(__file__).stem}_{getpass.getuser()}_{topic}'


TWO_ALG_ALG1 = base_port.next()
TWO_ALG_SHARE1 = base_port.next()
TWO_ALG_SHARE2 = base_port.next()

TWO_ALG_ALG1_DISTRIBUTED_CLAUSE = {
    'my_id': 1,
    'discoverer': {
        'name': 'static',
        'static': {
            'adjacency': {'1': []}
        }
    },
    'communicator': {
        'name': 'sockets',
        'sockets': {
            'nodes_and_endpoints': [
                ('1', (LOOPBACK, TWO_ALG_ALG1)),
            ],
        }
    },
}

TWO_ALG_ABSTRACTOR_DISTRIBUTED_CLAUSE = {
    'my_id': -1,
    'discoverer': {
        'name': 'static',
        'static': {
            'adjacency': {'1': [2], '2': [1]}
        }
    },
    'communicator': {
        'name': 'memory',
        'memory': {
            'domain': 'two_algorithms',
        }
    },
}

TWO_ALG_PDEF = {
    "dataset": {
        "input_format": ["sklearn.datasets.load_breast_cancer"],
        "loaded_format": ["pandas_dataframe"],
        "column_roles": {
            "target": {
                "name": "target"
            }
        },
        "params": {
            "test_size": 30,
        }
    },
    'problem_type': {
        'task': 'binary_classification'
    }
}


def test_two_algorithms(wait_til_all_fit: Waiter,
                        assert_no_exceptions: Clearer) -> None:
    '''Two different algorithms share data space.

    This is an integration test to show how to use an abstractor to share training between
    two disparate algorithms.

    This is intended as a model for integration tests for other abstractors.

    The ShareAbstractor just shares and concatenates data, so both Algorithms should see the same
    input data and thus make similar predictions.
    '''

    # The abstractor network has 2 nodes.
    distributed_clause_share1 = deepcopy(TWO_ALG_ABSTRACTOR_DISTRIBUTED_CLAUSE)
    distributed_clause_share1['my_id'] = 1
    distributed_clause_share2 = deepcopy(TWO_ALG_ABSTRACTOR_DISTRIBUTED_CLAUSE)
    distributed_clause_share2['my_id'] = 2

    wrangler = Wrangler(problem_definition=ProblemDefinition(TWO_ALG_PDEF))

    sklearn_svm_svc = wrangler.algorithm_catalog.lookup_by_name('sklearn.svm.SVC')

    dut_alg1 = AutonDecisionTreeModel().instantiate(
        distributed=DistributedConfig(clause=TWO_ALG_ALG1_DISTRIBUTED_CLAUSE),
        Lambda=10000000, omega=0.6)
    dut_share1 = ShareAbstractor().instantiate(
        distributed=DistributedConfig(clause=distributed_clause_share1),
        Lambda=10000000, omega=0.6)
    dut_share2 = ShareAbstractor().instantiate(
        distributed=DistributedConfig(clause=distributed_clause_share2),
        Lambda=10000000, omega=0.6)
    dut_alg2 = sklearn_svm_svc.instantiate(C=1.0)

    train = wrangler.load_train_dataset()
    assert train is not None
    train = train.sorted_columns()
    train_df = train.dataframe

    # Set up a pair of disjoint datasets.
    data_1_df = train_df.iloc[:270]
    data_2_df = train_df.iloc[270:]

    data_1 = train.output()
    data_1.dataframe = data_1_df

    data_2 = train.output()
    data_2.dataframe = data_2_df

    test = wrangler.load_test_dataset()
    assert test is not None
    test = test.sorted_columns()

    ground_truth = wrangler.load_ground_truth_dataset()
    assert ground_truth is not None
    ground_truth = ground_truth.sorted_columns()

    dut_alg1.start()
    dut_share1.start()
    dut_share2.start()
    try:
        dut_share1.fit(data_1)
        dut_share2.fit(data_2)

        wait_til_all_fit([dut_share1, dut_share2])

        dut_alg1.fit(dut_share1.synthesize())
        dut_alg2.fit(dut_share2.synthesize())
        wait_til_all_fit([dut_alg1])

    finally:
        dut_alg1.stop()
        dut_share1.stop()
        dut_share2.stop()

    assert_no_exceptions([dut_alg1, dut_share1, dut_share2])

    result1 = dut_alg1.predict(test)
    result2 = dut_alg2.predict(test)

    acc = wrangler.metric_catalog.lookup_by_name('accuracy_score')
    assert acc.calculate(result1, ground_truth) > 0.75
    assert acc.calculate(result2, ground_truth) > 0.75
    assert acc.calculate(result1, result2) > 0.73



'''Base module for catalogs in AutonML.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, Generic, Iterable, List, Set, Tuple, TypeVar, Union

T = TypeVar("T")


class CatalogError(Exception):
    '''Base error class for the catalog module'''


class CatalogLookupError(CatalogError, LookupError):
    '''The requested key is not present.'''


class CatalogNameError(CatalogError, LookupError):
    '''The object does not have a name attribute.'''


class CatalogValueError(CatalogError, ValueError):
    '''An improper value has been sent.'''


class Catalog(Generic[T], metaclass=abc.ABCMeta):
    '''Generic Catalog class for managing swappable components.

    Allows search by name of the component, and a lookup
    by classification or property.
    '''

    @property
    @abc.abstractmethod
    def tagtypes(self) -> Set[str]:
        '''Retrieve list of all tag types present in catalog'''

    @abc.abstractmethod
    def all_objects(self) -> Iterable[T]:
        '''Retrieve all registered objects.'''

    @abc.abstractmethod
    def items(self) -> Iterable[Tuple[str, T]]:
        '''List all the items in the catalog.'''

    @abc.abstractmethod
    def tagvals(self, tagtype: str) -> Set[str]:
        '''Retrieve all tag values present for a given tag type'''

    @abc.abstractmethod
    def lookup_by_name(self, name: str) -> T:
        '''Find an object registered under name.'''

    @abc.abstractmethod
    def lookup_by_tag_and(self, **tags: Union[str, Iterable[str]]) -> Dict[str, T]:
        '''Find objects with a specified tag for all specified tag types.

        \\*\\*tags looks like:

            tag_type = tag_value
            OR
            tag_type = (tag_value1, tag_value2)

        Needs to match at least one specified tag value for all specified tag types.
        '''


def upcast(tags: Dict[str, List[str]]) -> Dict[str, Union[str, List[str]]]:
    '''Cast a Dict[str, List[str]] into a Dict[str, Union[str, List[str]]]

    This is only needed to work within typing constraints.
    '''
    return dict(tags.items())



# flake8: noqa
# type: ignore
# pylint: skip-file

# ngautonml's Gaussian mixture implementation is adapted from this code made by Kyle Miller
# We are keeping it here for now to compare and make sure we adapted it right.
# There is a test that compares the ngautonml implementation against this code.

if __name__=="__main__":
    from DistAIModel import DistAIModel
else:
    from .DistAIModel import DistAIModel
import pickle
from typing import Any

import numpy as np
from scipy.linalg import solve_triangular
from scipy.special import logsumexp

from scipy.optimize import minimize, LinearConstraint, Bounds

def _logdet2d(cov_tril):
    return 2*np.log(np.diag(cov_tril)).sum()

def logdeterminants(cov_tril):
    if cov_tril.ndim==2: # diagonal covariances
        return np.log(cov_tril).sum(1)
    K = cov_tril.shape[0]
    return np.array([_logdet2d(cov_tril[k]) for k in range(K)])

def log_likelihoods(X,means,cov_tril):
    '''
    compute the log likelihood of observations X under each component of a GMM
    '''
    n,d = X.shape
    K = means.shape[0]
    Xmeaned = X[:,None,:]-means[None,:,:] # (n,.,d),(.,K,d)->(n,K,d)
    logdets = logdeterminants(cov_tril)
    if cov_tril.ndim==2: # diagonal covariances
        logprob = -0.5*np.einsum('kd,nkd->nk',1./cov_tril,Xmeaned**2) -0.5*logdets[None,:] -d*0.5*np.log(2*np.pi)
    else:
        z = np.stack([solve_triangular(cov_tril[k], Xmeaned[:,k,:].T, lower=True) for k in range(K)],2) # shape (d,n,K)
        logprob = -0.5*(z**2).sum(0) -0.5*logdets[None,:] -d*0.5*np.log(2*np.pi)
    return logprob # shape=(n,K)

def lp(X,mix,means,cov_tril):
    '''
    compute the log likelihood of observations X under a GMM
    '''
    logprob = log_likelihoods(X,means,cov_tril) # shape=(n,K)
    return np.apply_along_axis(logsumexp,1,logprob + np.log(mix)[None,:]).sum() # scalar

def grad_lp(X,mix,means,cov_tril,wrt_cov=None):
    '''
    compute the log likelihood of observations X under a GMM and gradients
      compute grad of cov_tril with respect to wrt_cov (i.e. sigma = L@L.T+M@M.T, and want gradient with respect to L)
         only needed for full covariances
    '''
    n,d = X.shape
    K = means.shape[0]
    Xmeaned = X[:,None,:]-means[None,:,:] # (n,.,d),(.,K,d)->(n,K,d)
    logdets = logdeterminants(cov_tril)
    if cov_tril.ndim==2: # diagonal covariances
        logprob = -0.5*np.einsum('kd,nkd->nk',1./cov_tril,Xmeaned**2) -0.5*logdets[None,:] -d*0.5*np.log(2*np.pi) # shape = (n,K)
        grad_means = Xmeaned/cov_tril[None,:,:] # shape = (n,K,d)
        grad_cov_tril = 0.5*(Xmeaned**2-cov_tril[None,:,:])/cov_tril[None,:,:]**2 # shape = (n,K,d)
    else:
        z = np.stack([solve_triangular(cov_tril[k], Xmeaned[:,k,:].T, lower=True) for k in range(K)],2) # shape (d,n,K)
        logprob = -0.5*(z**2).sum(0) -0.5*logdets[None,:] -d*0.5*np.log(2*np.pi) # shape (n,k)
        grad_means = np.stack([solve_triangular(cov_tril[k].T, z[:,:,k], lower=False) for k in range(K)],1).T # shape (n,K,d)
        grad_cov_tril = np.stack([np.einsum('ni,nj->nij',z[:,:,k].T,z[:,:,k].T) - np.eye(d)[None,:,:] for k in range(K)],1) # shape (n,K,d,d)
    #
    weights = logprob + np.log(mix)[None,:] # shape = (n,k)
    mx = weights.max(1)
    weights = np.exp( weights - mx[:,None] )
    wtot = weights.sum(1)
    wtotinv = 1./wtot
    #
    lp_ = np.log(wtot).sum()+mx.sum() # row-wise logsumexp -> scalar
    #
    grad_mix = np.einsum('nk,k,n->k',weights,1./mix,wtotinv) # shape = (k,)
    #
    grad_means = np.einsum('nk,nki,n->ki',weights,grad_means,wtotinv) # shape = (k,d)
    #
    if cov_tril.ndim==2: # diagonal covariances
        grad_cov_tril = np.einsum('nk,nki,n->ki',weights,grad_cov_tril,wtotinv)
    else:
        grad_cov_tril = np.einsum('nk,nkij,n->kij',weights,grad_cov_tril,wtotinv)
        if wrt_cov is not None:
            grad_cov_tril = np.stack([solve_triangular(cov_tril[k].T,grad_cov_tril[k]@solve_triangular(cov_tril[k],wrt_cov[k],lower=True),lower=False) for k in range(K)],0)
        else:
            grad_cov_tril = np.stack([solve_triangular(cov_tril[k].T,grad_cov_tril[k],lower=False) for k in range(K)],0)
        ri, ci = np.triu_indices(d,k=1)
        for k in range(K): grad_cov_tril[k,ri,ci] = 0
    #
    return lp_,grad_mix,grad_means,grad_cov_tril

def norm2(mix,means,cov_tril):
    return inner_product(mix,means,cov_tril,mix,means,cov_tril)

def grad_norm2(mix,means,cov_tril):
    norm2_, grad_mix, grad_means, grad_cov_tril = grad_inner_product(mix,means,cov_tril,mix,means,cov_tril)
    return norm2_, 2*grad_mix, 2*grad_means, 2*grad_cov_tril

def inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril):
    K,d = means.shape
    inner_product_ = 0.
    for i in range(K):
        for j in range(K):
            if cov_tril.ndim==2: # diagonal covariances
                cov = cov_tril[i]+other_cov_tril[j]
                z = (means[i]-other_means[j])/np.sqrt(cov)
                logdet = np.log(cov).sum()
            else:
                L = np.linalg.cholesky(cov_tril[i]@cov_tril[i].T+other_cov_tril[j]@other_cov_tril[j].T)
                z = solve_triangular(L,other_means[j]-means[i],lower=True)
                logdet = _logdet2d(L)
            inner_product_ += mix[i]*other_mix[j]*np.exp( -0.5*z@z -0.5*logdet -d*0.5*np.log(2*np.pi) )
    return inner_product_

def grad_inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril):
    K,d = means.shape
    inner_product_ = 0.
    grad_mix = np.zeros(K)
    grad_means = np.zeros((K,d))
    grad_cov_tril = np.zeros(cov_tril.shape)
    for j in range(K):
        if cov_tril.ndim==2: # diagonal covariances
            cov = cov_tril+other_cov_tril[j][None,:]
        else:
            cov = np.stack([np.linalg.cholesky(cov_tril[i]@cov_tril[i].T+other_cov_tril[j]@other_cov_tril[j].T) for i in range(K)],0)
        lp_, grad_mix_, grad_means_, grad_cov_tril_ = grad_lp(other_means[j][None,:],mix,means,cov,cov_tril)
        p = np.exp(lp_)
        inner_product_ += other_mix[j]*p
        grad_mix += other_mix[j]*p*grad_mix_
        grad_means += other_mix[j]*p*grad_means_
        grad_cov_tril += other_mix[j]*p*grad_cov_tril_
    return inner_product_, grad_mix, grad_means, grad_cov_tril

#################################################################
#################################################################
# Gaussian mixture model

class GMM(DistAIModel):

    def __init__(self, config):
        super(GMM, self).__init__()
        if config is None: return
        self.K = int( config['K'] )
        self.full_covariance = config['covariance'].lower()=='full'
        self.Lambda = float( config['lambda'] )
        self.omega = float( config['omega'] )
        self.fit_time = None
        self.ID = None
        self.mix = None
        self.means = None
        self.cov_tril = None
        self.random_state = np.random.RandomState(1701)

    def serialize(self): # returns serialized model
        # care should be taken to make message size as small as possible. self.ID is required.
        # the deserialized model must be able to run norm2, inner_product, and metrics. It need not run fit.
        if self.full_covariance:
            d = self.means.shape[1]
            cov = np.stack([self.cov_tril[k][np.tril_indices(d)] for k in range(self.K)],0)
        else:
            cov = self.cov_tril
        return pickle.dumps((self.ID,self.mix,self.means,cov))

    @staticmethod
    def deserialize(serialized_model): # returns model instance
        # care should be taken to make message size as small as possible. self.ID is required.
        # the deserialized model must be able to run norm2, inner_product, and metrics. It need not run fit.
        obj = pickle.loads(serialized_model)
        m = GMM(None)
        m.ID = obj[0]
        m.mix = obj[1]
        m.means = obj[2]
        K,d = m.means.shape
        m.K = K
        cov = obj[3]
        if cov.shape[1] > m.means.shape[1]:
            m.full_covariance = True
            m.cov_tril = np.zeros((K,d,d))
            for k in range(K):
                m.cov_tril[k][np.tril_indices(d)] = cov[k]
        else:
            m.full_covariance = False
            m.cov_tril = cov
        return m

    def sample(self,n):
        # sample
        pass

    def lp(self,X): # log_probability of X. X shape=(n,d)
        '''
        Log-likelihood of the data
        '''
        return lp(X,self.mix,self.means,self.cov_tril) #scalar

    def idx(self,X): # log_probability of X. X shape=(n,d)
        '''
        Maximum likelihood component index for each observation
        '''
        ll = log_likelihoods(X,self.means,self.cov_tril) # shape = (n,K)
        return np.argmax( ll+np.log(self.mix)[None,:], 1)

    def p(self,X):
        '''
        likelihood of data
        '''
        return np.exp(self.lp(X)) #scalar

    @staticmethod
    def flatten(mix,means,cov_tril):
        if cov_tril.ndim==2:
            return np.concatenate((mix,means.flatten(),cov_tril.flatten()))
        K,d = means.shape
        idx = np.tril_indices(d)
        return np.concatenate([mix,means.flatten()]+[cov_tril[k][idx] for k in range(K)])

    @staticmethod
    def unflatten(x,K,d,full):
        if not full:
            return x[:K], x[K:(K+K*d)].reshape((K,d)), x[(K+K*d):].reshape((K,d))
        cov = x[(K+K*d):]
        cov_tril = np.zeros((K,d,d))
        s = (d*(d+1))//2
        idx = np.tril_indices(d)
        for k in range(K):
            cov_tril[k][idx] = cov[(k*s):((k+1)*s)]
        return x[:K], x[K:(K+K*d)].reshape((K,d)), cov_tril

    def fit(self, data, neighbor_models): # returns nothing, fits model to data and other models
        d = data.shape[1]
        K = self.K
        # initial setup
        last_fit = None
        Nneighbors = len(neighbor_models)
        if self.means is None: # first fit
            self.means = self.random_state.rand(K,d)*(data.max(0)-data.min(0)+1e-6)[None,:] + data.min(0) if data.shape[0]>0 else self.random_state.rand(K,d)-0.5
            self.mix = np.ones(K)/K
            if self.full_covariance:
                self.cov_tril = np.zeros((K,d,d)) + np.eye(d)[None,:,:]
            else:
                self.cov_tril = np.ones((K,d))
        else:
            last_fit = GMM.deserialize( self.serialize() )
        # objective
        x0 = GMM.flatten(self.mix,self.means,self.cov_tril)
        def fun(x):
            mix, means, cov_tril = GMM.unflatten(x,K,d,self.full_covariance)
            JacMix = np.eye(mix.size)/mix.sum()-mix[None,:]/mix.sum()**2
            mix = mix/mix.sum()
            # local data
            lp_, grad_mix, grad_means, grad_cov_tril = grad_lp(data,mix,means,cov_tril)
            grad = GMM.flatten(JacMix@grad_mix,grad_means,grad_cov_tril)
            if last_fit is None:
                return -lp_, -grad
            # regularization
            reg = 0
            reg_grad = np.zeros(x.size)
            nrm2, grad_mix, grad_means, grad_cov_tril = grad_norm2(mix,means,cov_tril)
            reg += 1./self.omega*nrm2
            reg_grad += 1./self.omega*GMM.flatten(JacMix@grad_mix,grad_means,grad_cov_tril)
            ip, grad_mix, grad_means, grad_cov_tril = grad_inner_product(mix,means,cov_tril,last_fit.mix,last_fit.means,last_fit.cov_tril)
            reg += 2.*(1.-1./self.omega)*ip
            reg_grad += 2.*(1.-1./self.omega)*GMM.flatten(JacMix@grad_mix,grad_means,grad_cov_tril)
            for n in range(Nneighbors):
                ip, grad_mix, grad_means, grad_cov_tril = grad_inner_product(mix,means,cov_tril,neighbor_models[n].mix,neighbor_models[n].means,neighbor_models[n].cov_tril)
                reg += -2./Nneighbors*ip
                reg_grad += -2./Nneighbors*GMM.flatten(JacMix@grad_mix,grad_means,grad_cov_tril)
            #
            return -lp_+self.Lambda*reg, -grad+self.Lambda*reg_grad
        lb = (-np.inf)*np.ones(x0.size)
        ub = (np.inf)*np.ones(x0.size)
        lb[:K] = 1e-6
        ub[:K] = 1.
        if not self.full_covariance:
            lb[(K+K*d):] = 1e-4
        else:
            ri,ci=np.tril_indices(d)
            lb[(K+K*d)+np.where(np.tile(ri==ci,K))[0]] = 1e-2 # constrain diagonal of cov_tril to be positive
        bnds = Bounds(lb=lb,ub=ub,keep_feasible=True)
        soln = minimize(fun, x0, method='L-BFGS-B', jac=True, bounds=bnds)
        self.mix, self.means, self.cov_tril = GMM.unflatten(soln.x,K,d,self.full_covariance)
        self.mix = self.mix/self.mix.sum()

    def norm2(self): # returns scalar
        return norm2(self.mix,self.means,self.cov_tril)

    def inner_product(self,other): # returns scalar
        return inner_product(self.mix,self.means,self.cov_tril,other.mix,other.means,other.cov_tril)

    def metrics(self,data): # returns a dictionary with string keys recording any sort of metrics Dict[str, Any]
        return {'LogProb':self.lp(data)/data.shape[0]} # mean log probability

#################################################################
#################################################################

if __name__=="__main__":
    '''
    Do gradient checks:
    NOTE:  d/dA f(Ainv) = -Ainv.T @ df/dAinv @ Ainv.T

    dlogrpob( x | mu, sigma )/dmu = inv(sigma)(x-mu)

    dlogprob( x | mu, sigma )/dsigma_inv = -0.5(x-mu)(x-mu).T + 0.5*sigma
    dlogprob( x | mu, sigma )/dsigma = -inv(sigma).T@(-0.5(x-mu)(x-mu).T + 0.5*sigma)@inv(sigma).T
    dlogprob( x | mu, L@L.T + M@M.T )/dL = 2*(-inv(sigma).T@(-0.5(x-mu)(x-mu).T + 0.5*sigma)@inv(sigma).T)@L
                                         = inv(R.T)@( (inv(R)(x-mu))(inv(R)(x-mu)).T - I )@inv(R)@L   where R=cholesky(sigma)
    '''
    seed = 0
    np.random.seed(seed)
    full_cov = False
    eps = 1e-4
    n,d,K = 10,2,3
    X = np.random.random((n,d))
    #
    if not full_cov:
        cov_tril = np.random.random((K,d))
    else:
        cov_tril = np.random.random((K,d,d))
        for k in range(K): cov_tril[k] = np.linalg.cholesky( cov_tril[k]@cov_tril[k].T )
    #
    means = np.random.random((K,d))
    mix = np.random.random(K)
    mix = mix/mix.sum()
    def chk(X,mix,means,cov_tril):
        return grad_lp(X,mix,means,cov_tril)[0]
    #
    print("lp")
    print(lp(X,mix,means,cov_tril),grad_lp(X,mix,means,cov_tril)[0])
    #grad check mix
    print("mix lp")
    grad = np.zeros(K)
    for i in range(K):
        mix_p = mix.copy()
        mix_p[i] += eps
        mix_m = mix.copy()
        mix_m[i] -= eps
        grad[i] = (chk(X,mix_p,means,cov_tril)-chk(X,mix_m,means,cov_tril))/(2*eps)
    print( np.abs(grad-grad_lp(X,mix,means,cov_tril)[1]).max(), np.abs((grad-grad_lp(X,mix,means,cov_tril)[1])/grad).max() )
    #
    #grad check means
    print("mean lp")
    grad = np.zeros((K,d))
    for i in range(K):
        for j in range(d):
            means_p = means.copy()
            means_p[i,j] += eps
            means_m = means.copy()
            means_m[i,j] -= eps
            grad[i,j] = (chk(X,mix,means_p,cov_tril)-chk(X,mix,means_m,cov_tril))/(2*eps)
    print( np.abs(grad-grad_lp(X,mix,means,cov_tril)[2]).max(), np.abs((grad-grad_lp(X,mix,means,cov_tril)[2])/grad).max() )
    #
    #grad check cov
    print("cov lp")
    grad = np.zeros(cov_tril.shape)
    for k in range(K):
        for i in range(d):
            if not full_cov:
                cov_tril_p = cov_tril.copy()
                cov_tril_p[k,i] += eps
                cov_tril_m = cov_tril.copy()
                cov_tril_m[k,i] -= eps
                grad[k,i] = (chk(X,mix,means,cov_tril_p)-chk(X,mix,means,cov_tril_m))/(2*eps)
            else:
                for j in range(i+1):
                    cov_tril_p = cov_tril.copy()
                    cov_tril_p[k,i,j] += eps
                    cov_tril_m = cov_tril.copy()
                    cov_tril_m[k,i,j] -= eps
                    grad[k,i,j] = (chk(X,mix,means,cov_tril_p)-chk(X,mix,means,cov_tril_m))/(2*eps)
    print( np.abs(grad-grad_lp(X,mix,means,cov_tril)[3]).max(), np.abs((grad-grad_lp(X,mix,means,cov_tril)[3])/(grad+1e-60)).max() )
    #######
    def chk(mix,means,cov_tril):
        return grad_norm2(mix,means,cov_tril)[0]
    #
    print("norm2")
    print(norm2(mix,means,cov_tril),grad_norm2(mix,means,cov_tril)[0])
    #grad check mix
    print("mix norm2")
    grad = np.zeros(K)
    for i in range(K):
        mix_p = mix.copy()
        mix_p[i] += eps
        mix_m = mix.copy()
        mix_m[i] -= eps
        grad[i] = (chk(mix_p,means,cov_tril)-chk(mix_m,means,cov_tril))/(2*eps)
    print( np.abs(grad-grad_norm2(mix,means,cov_tril)[1]).max(), np.abs((grad-grad_norm2(mix,means,cov_tril)[1])/grad).max() )
    #
    #grad check means
    print("mean norm2")
    grad = np.zeros((K,d))
    for i in range(K):
        for j in range(d):
            means_p = means.copy()
            means_p[i,j] += eps
            means_m = means.copy()
            means_m[i,j] -= eps
            grad[i,j] = (chk(mix,means_p,cov_tril)-chk(mix,means_m,cov_tril))/(2*eps)
    print( np.abs(grad-grad_norm2(mix,means,cov_tril)[2]).max(), np.abs((grad-grad_norm2(mix,means,cov_tril)[2])/grad).max() )
    #
    #grad check cov
    print("cov norm2")
    grad = np.zeros(cov_tril.shape)
    for k in range(K):
        for i in range(d):
            if not full_cov:
                cov_tril_p = cov_tril.copy()
                cov_tril_p[k,i] += eps
                cov_tril_m = cov_tril.copy()
                cov_tril_m[k,i] -= eps
                grad[k,i] = (chk(mix,means,cov_tril_p)-chk(mix,means,cov_tril_m))/(2*eps)
            else:
                for j in range(i+1):
                    cov_tril_p = cov_tril.copy()
                    cov_tril_p[k,i,j] += eps
                    cov_tril_m = cov_tril.copy()
                    cov_tril_m[k,i,j] -= eps
                    grad[k,i,j] = (chk(mix,means,cov_tril_p)-chk(mix,means,cov_tril_m))/(2*eps)
    print( np.abs(grad-grad_norm2(mix,means,cov_tril)[3]).max(), np.abs((grad-grad_norm2(mix,means,cov_tril)[3])/(grad+1e-60)).max() )
    #######
    def chk(mix,means,cov_tril):
        return grad_inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril)[0]
    #
    print("inner product")
    if not full_cov:
        other_cov_tril = np.random.random((K,d))
    else:
        other_cov_tril = np.random.random((K,d,d))
        for k in range(K): other_cov_tril[k] = np.linalg.cholesky( other_cov_tril[k]@other_cov_tril[k].T )
    #
    other_means = np.random.random((K,d))
    other_mix = np.random.random(K)
    other_mix = other_mix/other_mix.sum()
    print(inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril),grad_inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril)[0])
    #grad check mix
    print("mix inner product")
    grad = np.zeros(K)
    for i in range(K):
        mix_p = mix.copy()
        mix_p[i] += eps
        mix_m = mix.copy()
        mix_m[i] -= eps
        grad[i] = (chk(mix_p,means,cov_tril)-chk(mix_m,means,cov_tril))/(2*eps)
    print( np.abs(grad-grad_inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril)[1]).max(), np.abs((grad-grad_inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril)[1])/grad).max() )
    #
    #grad check means
    print("mean inner product")
    grad = np.zeros((K,d))
    for i in range(K):
        for j in range(d):
            means_p = means.copy()
            means_p[i,j] += eps
            means_m = means.copy()
            means_m[i,j] -= eps
            grad[i,j] = (chk(mix,means_p,cov_tril)-chk(mix,means_m,cov_tril))/(2*eps)
    print( np.abs(grad-grad_inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril)[2]).max(), np.abs((grad-grad_inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril)[2])/grad).max() )
    #
    #grad check cov
    print("cov inner product")
    grad = np.zeros(cov_tril.shape)
    for k in range(K):
        for i in range(d):
            if not full_cov:
                cov_tril_p = cov_tril.copy()
                cov_tril_p[k,i] += eps
                cov_tril_m = cov_tril.copy()
                cov_tril_m[k,i] -= eps
                grad[k,i] = (chk(mix,means,cov_tril_p)-chk(mix,means,cov_tril_m))/(2*eps)
            else:
                for j in range(i+1):
                    cov_tril_p = cov_tril.copy()
                    cov_tril_p[k,i,j] += eps
                    cov_tril_m = cov_tril.copy()
                    cov_tril_m[k,i,j] -= eps
                    grad[k,i,j] = (chk(mix,means,cov_tril_p)-chk(mix,means,cov_tril_m))/(2*eps)
    print( np.abs(grad-grad_inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril)[3]).max(), np.abs((grad-grad_inner_product(mix,means,cov_tril,other_mix,other_means,other_cov_tril)[3])/(grad+1e-60)).max() )

    #####################################
    import matplotlib.pyplot as plt

    # test GMM
    n = [50,25,40]
    mu = [[-1,0],[0.2,-0.2],[0,0.5]]
    cov = [[[0.025,0],[0,0.025]],[[0.025,0.0125],[0.0125,0.025]],[[0.05,-0.01],[-0.01,0.05]]]
    rng = np.random.default_rng(seed=seed)
    X = []
    for i in range(len(n)):
        X.append( rng.multivariate_normal(mu[i],cov[i],n[i]) )

    X = np.concatenate(X)
    plt.scatter(X[:,0],X[:,1])
    plt.show()

    config = {'K':3,'covariance':'full','lambda':1.,'omega':0.667}

    model = GMM(config)

    model.fit(X,[])
    idx = model.idx(X)
    for i in range(config['K']):
        plt.scatter(X[idx==i,0],X[idx==i,1])
        plt.scatter(model.means[i,0],model.means[i,1],marker='^')
    plt.show()
    print(model.means)
    print(model.cov_tril)
    print(model.mix)

    model.fit(X,[])
    idx = model.idx(X)
    for i in range(config['K']):
        plt.scatter(X[idx==i,0],X[idx==i,1])
        plt.scatter(model.means[i,0],model.means[i,1],marker='^')
    plt.show()
    print(model.means)
    print(model.cov_tril)
    print(model.mix)

def register(catalog: Any, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''


'''Tests for SimpleInstantiator'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import Designator
from ...algorithms.connect import ConnectorModel
from ...wrangler.dataset import Dataset

from .simple_instantiator import SimpleInstantiator

# pylint: disable=missing-function-docstring,duplicate-code


REFERENCE_DATASET = Dataset(
    a_key='a_value',
    another_key='another_value',
)


def test_sunny_day() -> None:
    pipe = BoundPipeline(name=Designator('foo'), tags={})
    connector = ConnectorModel()
    pipe.step(model=connector, new_key='a_key', another_key='another_key').set_name('connector1')
    pipe.step(model=connector,
              twice_transformed_key='new_key', another_key='another_key').set_name('connector2')

    dut = SimpleInstantiator()

    executable = dut.instantiate(pipeline=pipe)

    executable.fit(dataset=REFERENCE_DATASET)

    got_predict = executable.predict(dataset=REFERENCE_DATASET)
    want_predict = {
        'twice_transformed_key': 'a_value',
        'another_key': 'another_value',
    }
    assert got_predict.prediction is not None
    assert set(got_predict.prediction) == set(want_predict)



'''Tests replicating example notebooks.'''
# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
from pathlib import Path

from ..algorithms.impl.algorithm_auto import FakeAlgorithmCatalogAuto
from ..problem_def.problem_def import ProblemDefinition
from ..tables.impl.table_auto import TableCatalogAuto
from .wrangler import Wrangler

_ = TableCatalogAuto()


def module_path() -> Path:
    current_path = str(os.getenv('PYTEST_CURRENT_TEST')).split('::', maxsplit=1)[0]
    pathobj = Path(current_path).resolve()
    module_parent = pathobj.parents[2]
    return module_parent


def credit_problem_def():
    train_path = module_path() / 'examples' / 'classification' / 'credit-train.csv'
    pdef = {
        "dataset": {
            "input_format": "tabular_file",
            "loaded_format": "pandas_dataframe",
            "column_roles": {
                "target": {
                    "name": "class"
                }
            },
            "params": {
                "train_path": train_path,
            }
        },
        "problem_type": {
            "task": "binary_classification"
        },
        "cross_validation": {
            "k": 2
        },
        "metrics": {
            "accuracy_score": {},
            "roc_auc_score": {}
        }
    }
    return ProblemDefinition(pdef)


def test_wrangle_credit() -> None:
    problem_def = credit_problem_def()
    dut = Wrangler(
        problem_definition=problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)
    got = dut.fit_predict_rank()

    assert got.split_dataset.ground_truth is not None
    assert len(got.split_dataset.ground_truth.ground_truth) == 800
    got_train_result = list(got.train_results.values())[0].prediction
    assert got_train_result is not None
    assert got_train_result.predictions.shape == (800, 2)

    # For accuracy, we can't guarantee which pipe will be better, but
    #   we can check that both are present.
    assert 'accuracy_score' in got.rankings
    pipes_included = {
        str(score.result.bound_pipeline.designator)
        for score in got.rankings['accuracy_score'].scores(all_scores=True)}
    assert len(pipes_included) == 36
    assert {
        'tabular_classification@sklearn.linear_model.logisticregression:c=100.0,class_weight=none',
        'tabular_classification@sklearn.svm.svc:c=0.01,class_weight=none'
    }.issubset(pipes_included)

    # For roc_auc, we know logistic regression is scorable and svc is unscorable
    #   (because svc cannot produce probabilities output.)
    # As a result, svc will always be ranked last and its score will be
    #   a str stating that it cannot be scored.
    assert 'roc_auc_score' in got.rankings
    best_score = got.rankings['roc_auc_score'].best(1, all_scores=True)[0]
    best_pipe = str(best_score.pipeline_des)
    # Small changes in the implementation of libraries seems to have caused the order
    # of these two pipelines to change. We're happy if either of these is the top pipeline.
    assert (
        best_pipe.startswith(
            'method_of_moments@sklearn.linear_model.logisticregression:')
        or best_pipe.startswith(
            'tabular_classification@sklearn.linear_model.logisticregression:')
    )

    assert isinstance(best_score.score, float)

    worst_score = got.rankings['roc_auc_score'].scores(all_scores=True)[-1]
    assert 'sklearn.svm.svc' in str(worst_score.pipeline_des)
    assert isinstance(worst_score.score, str)  # It is an error message, not a score.


def iris_problem_def():
    train_path = module_path() / 'examples' / 'iris' / 'iris.csv'
    pdef = {
        "dataset": {
            "config": "local",
            "train_path": train_path,
            "column_roles": {
                "target": {
                    "name": "flowers"
                }
            }
        },
        "problem_type": {
            "task": "multiclass_classification"
        },
        "cross_validation": {
            "k": 2
        }
    }
    return ProblemDefinition(pdef)


def test_wrangle_iris() -> None:
    problem_def = iris_problem_def()
    dut = Wrangler(
        problem_definition=problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)
    got = dut.fit_predict_rank()

    assert got.split_dataset.ground_truth is not None
    assert len(got.split_dataset.ground_truth.ground_truth) == 150
    got_train_results = list(got.train_results.values())[0].prediction
    assert got_train_results is not None
    assert got_train_results.predictions.shape == (150, 2)

    # For accuracy, we can't guarantee which pipe will be better, but
    #   we can check that both are present.
    # We know accuracy will be the first Ranking because they are sorted
    #   alphabetically by metric name.
    pipes_included = {
        str(score.result.bound_pipeline.designator)
        for score in got.rankings['accuracy_score'].scores(all_scores=True)}
    assert len(pipes_included) == 18
    assert {
        'tabular_classification@sklearn.linear_model.logisticregression:c=100.0,class_weight=none',
        'tabular_classification@sklearn.svm.svc:c=0.01,class_weight=none'
    }.issubset(pipes_included)

    # TODO(Merritt): make roc_auc_score work for multiclass


def diabetes_config():
    train_path = module_path() / 'examples' / 'regression' / 'diabetes.csv'
    pdef = {
        "dataset": {
            "config": "local",
            "train_path": train_path,
            "column_roles": {
                "target": {
                    "name": "Prog"
                }
            }
        },
        "problem_type": {
            "data_type": "TABULAR",
            "task": "REGRESSION"
        },
        "metrics": {
            "root_mean_squared_error": {}
        },
        "cross_validation": {
            "k": 2
        }
    }
    return ProblemDefinition(pdef)


def test_wrangle_diabetes() -> None:
    problem_definition = diabetes_config()
    dut = Wrangler(
        problem_definition=problem_definition,
        algorithm_catalog=FakeAlgorithmCatalogAuto
    )
    got = dut.fit_predict_rank()

    # there are 442 rows in examples/regression/diabetes.csv
    # we expect train predictions that cover all of them.
    got_train_results = list(got.train_results.values())[0].prediction
    assert got_train_results is not None
    assert got_train_results.predictions.shape == (442, 2)



'''Load JSON from disk into a pipeline.'''
import json
from pathlib import Path
from typing import Any, Dict, List, Optional

from .instantiator import InstantiatorError
from .loader import Loader
from ..algorithms.impl.algorithm import AlgorithmCatalog
from ..executor.cucumber import Cucumber
from ..generator.bound_pipeline import BoundPipeline
from ..templates.impl.pipeline_step import PipelineStep


class JsonLoader(Loader):
    '''Load JSON from disk into a pipeline.'''
    _pipeline: BoundPipeline
    _model_folder: Optional[Path]
    _algorithm_catalog: AlgorithmCatalog
    _load_models: bool

    def __init__(self,
                 saver_version: str,
                 algorithm_catalog: AlgorithmCatalog,
                 pipeline_file: Path,
                 model_folder: Optional[Path] = None,
                 load_models: bool = False):
        with pipeline_file.open('rb') as file_pointer:
            pipe = json.load(file_pointer)
        if saver_version != pipe['version']:
            raise InstantiatorError(
                f'This version of ngAutonML only supports {saver_version}, not {pipe["version"]}')

        self._algorithm_catalog = algorithm_catalog
        # TODO(Merritt/Piggy): replace this with a path manager
        self._model_folder = model_folder or (
            Path(pipe['output_dir']) / 'models' / pipe['pipeline_designator'])
        self._load_models = load_models
        pipeline = BoundPipeline(name=pipe['pipeline_template_name'])
        self._pipeline = self._load_pipeline(pipeline=pipeline, steps=pipe['steps'])

    @property
    def pipeline(self) -> BoundPipeline:
        '''Return the underlying pipeline.

        This is the main way of accessing the result of this object.
        '''
        return self._pipeline

    def _load_cucumber(self, step: Dict[str, Any]) -> Cucumber:
        model_path = self._model_folder / (step['model_filename'] + '.pkl')
        with model_path.open('rb') as file_pointer:
            return Cucumber.deserialize(file_pointer.read())

    def _load_step(self, pipeline: BoundPipeline, step: Dict[str, Any]) -> PipelineStep:
        assert self._algorithm_catalog is not None, (
            'BUG: We should not be loading without an algorithm catalog.'
        )
        hyperparams = step.get('hyperparams', {}).copy()
        if self._load_models:
            cucumber = self._load_cucumber(step)
            hyperparams.update(cucumber.hyperparams)
        assert isinstance(step['algorithm_name'], str)
        algorithm = self._algorithm_catalog.lookup_by_name(name=step['algorithm_name'])
        retval = pipeline.step(
            model=algorithm,
            serialized_model=cucumber.serialized_model,
            **hyperparams)
        if 'name' in step:
            retval.set_name(step['name'])
        return retval

    def _load_parallel(self,
                       pipeline: BoundPipeline,
                       pipelines: Dict[str, List[Dict[str, Any]]]) -> PipelineStep:
        subpipes = {}
        for subpipe, substeps in pipelines.items():
            newpipe = pipeline.new(name=subpipe)
            subpipes[subpipe] = self._load_pipeline(pipeline=newpipe, steps=substeps)
        return pipeline.parallel(**subpipes)

    def _load_pipeline(self, pipeline: BoundPipeline, steps: List[Dict[str, Any]]) -> BoundPipeline:
        retval = pipeline
        for step in steps:
            if step['type'] == 'parallel':
                self._load_parallel(pipeline=retval, pipelines=step['subpipelines'])
            elif step['type'] == 'step':
                self._load_step(pipeline=retval, step=step)
        return retval



'''A Table representing a dictionary from str to list-like.'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd

from .impl.table import TableFactory, Table
from .impl.table_catalog import TableCatalog


class DictTable(Table):
    '''A Table representing a dictionary from str to list-like.'''
    name = 'dict'
    _value: Dict[Union[str, int], List]
    my_type = dict

    def to_csv(self, *args, **kwargs) -> str:
        '''Output the table as a CSV string.'''
        return pd.DataFrame(self._value).to_csv(*args, **kwargs)

    def as_dict(self, *args, **kwargs) -> Dict[Union[str, int], Any]:
        '''Convert table to a dictionary.'''
        return self._value

    def drop(self, columns: List[Union[int, str]]) -> Table:
        '''Return a copy of this table with selected columns dropped.'''
        return TableFactory(
            {k: v for k, v in self._value.items()
             if k not in columns})

    @property
    def empty(self) -> bool:
        '''Return True if the table is empty.'''
        return len(self._value) == 0 or all(
            len(v) == 0 for v in self._value.values()
        )

    @property
    def shape(self) -> tuple:
        '''Get the shape of the table.'''
        return (len(next(iter(self._value.values()))), len(self.columns))

    @property
    def columns(self) -> List[Union[int, str]]:
        '''Return the column names of the table.'''
        return sorted(list(self._value.keys()))

    @columns.setter
    def columns(self, value: List[Union[int, str]]) -> None:
        '''Set the column names of the table.'''
        # TODO(Merritt/Piggy): We can get a sorted list of the current column names,
        # then iterate over the new column names and update the dictionary.
        raise NotImplementedError('Cannot set columns for a DictTable')


def as_dataframe(from_table: Table) -> pd.DataFrame:
    '''Export a DictTable to a pandas DataFrame.'''
    assert isinstance(from_table, DictTable)
    retval = pd.DataFrame(from_table.value())
    # Make sure columns are in a predictable order.
    return retval.reindex(from_table.columns, axis=1)


def as_numpy(from_table: Table) -> np.ndarray:
    '''Export a DictTable to a numpy array.'''
    df_table = as_dataframe(from_table)
    if df_table.shape[1] == 1:
        return df_table.to_numpy().flatten()
    return df_table.to_numpy()


def register(catalog: TableCatalog):
    '''Register the DataFrameTable with the catalog.'''
    catalog.register(DictTable)
    TableFactory.register_constructor(dict, DictTable)
    TableFactory.register_exporter(
        from_type=dict, to_type=dict, exporter=DictTable.value)
    TableFactory.register_exporter(
        from_type=dict, to_type=pd.DataFrame, exporter=as_dataframe)
    TableFactory.register_exporter(
        from_type=dict, to_type=np.ndarray, exporter=as_numpy)



'''Searcher object.

This object sits between the Generator and the Instantiator.
It implements static searching algoritms (like grid search)
and does hyperparameter overriding (a trivial special case
of search).
'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
import logging
from typing import Any, Dict, List, Optional, Set

from frozendict import frozendict

from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..problem_def.hyperparam_config import HyperparamConfig
from ..templates.impl.pipeline_step import PipelineStep
from ..wrangler.constants import RangeMethod

from .frozen_overrides import FrozenOverrides
from .params import Overrides, ParamRanges
from .param_range import ParamRange


class Searcher(metaclass=abc.ABCMeta):
    '''Generate new BoundPipelines based on hyperparam search spaces.'''
    _overrides: Overrides
    _disable_grid_search: bool
    _inject: Optional[Dict[str, Any]]

    def __init__(
        self,
        hyperparams: HyperparamConfig,
        inject: Optional[Overrides] = None
    ):
        self._overrides = hyperparams.overrides.copy()
        self._overrides.extend(inject or [])
        self._disable_grid_search = hyperparams.disable_grid_search

    @abc.abstractmethod
    def bind_hyperparams(self, pipeline: BoundPipeline) -> Dict[Designator, BoundPipeline]:
        '''Bind new BoundPipelines with revised hyperparameters.'''

    def bind_all(self,
                 pipelines: Dict[Designator, BoundPipeline]) -> Dict[Designator, BoundPipeline]:
        '''Bind all the hyperparams for a set of BoundPipelines.'''
        retval: Dict[Designator, BoundPipeline] = {}
        for pipe in pipelines.values():
            retval.update(self.bind_hyperparams(pipe))
        return retval


class TrivialSearcher(Searcher):
    '''Just return the pipeline passed to us. Not really a searcher.'''
    def bind_hyperparams(self, pipeline: BoundPipeline) -> Dict[Designator, BoundPipeline]:
        return {
            pipeline.designator: pipeline
        }


class StepSet():
    '''Turn a step into a set of steps for all hyperparam bindings.'''

    # The key of _steps is optional because at initialization we
    # need to use None as a key.
    _steps: Dict[Optional[frozendict[str, str]], PipelineStep]
    _algorithm: Any  # Breaking circularity; really an Algorithm
    _pipeline_designator_component: str

    def __init__(self, step: PipelineStep):
        self._steps = {None: step}
        assert step.has_model(), (
            'BUG: attempt to bind hyperparams for step '
            f'({step.pipeline_designator_component}) with no algorithm')
        self._algorithm = step.model
        self._pipeline_designator_component = step.pipeline_designator_component

    def multiply(self,
                 hyperparam_name: str,
                 param_range: ParamRange,
                 no_show: Set[str]) -> 'StepSet':
        '''Expand the set of steps by the number of values in param.'''
        new_steps: Dict[Optional[frozendict[str, str]], PipelineStep] = {}
        for old_bindings, old_step in self._steps.items():
            bindings: Dict[str, Any] = self._algorithm.bind(
                hyperparam_name=hyperparam_name, param_range=param_range)
            for new_value_str, new_value in bindings.items():

                new_step = old_step.clone(**{hyperparam_name: new_value})
                assert new_step.filename is not None

                new_bindings: Optional[frozendict[str, str]] = frozendict()
                if hyperparam_name in no_show:
                    if old_bindings is not None:
                        new_bindings = old_bindings
                    else:
                        new_bindings = None
                else:
                    if old_bindings is None:
                        # step does not have any pre-existing overrides
                        new_bindings = frozendict({hyperparam_name: new_value_str})
                    else:
                        # step has preexisting overrides
                        new_bindings = old_bindings.set(hyperparam_name, new_value_str)  # type: ignore[attr-defined]  # pylint: disable=line-too-long

                new_steps[new_bindings] = new_step
        self._steps = new_steps
        return self

    @property
    def steps(self) -> Dict[Optional[FrozenOverrides], PipelineStep]:
        '''Render self as a dict of PipelineSteps.'''
        retval: Dict[Optional[FrozenOverrides], PipelineStep] = {}
        for bind, step in self._steps.items():
            if bind is None:
                retval[None] = step
                continue
            retval[FrozenOverrides({
                self._pipeline_designator_component: bind
            })] = step
        return retval


class PipelineBuilder():
    '''Build a set of pipelines by repeatedly adding a single Step or a StepSet.'''
    _pipename: str
    _steps: Optional[Dict[Optional[FrozenOverrides], List[PipelineStep]]]

    def __init__(self, pipename: str):
        self._pipename = pipename
        self._steps = None

    def append(self, step_set: StepSet) -> 'PipelineBuilder':
        '''Expand self._steps, adding each step from step_set to the end of each step list.

        If len(self._steps) is n, and len(step_set) is m, len(self._steps) will become m*n.
        '''
        if self._steps is None:
            self._steps = {key: [step] for key, step in step_set.steps.items()}
            return self
        new_steps: Dict[Optional[FrozenOverrides], List[PipelineStep]] = {}
        for old_frozen_overrides, old_steps in self._steps.items():
            for new_frozen_overrides, new_step in step_set.steps.items():

                if new_frozen_overrides is not None and old_frozen_overrides is not None:
                    updated_frozen_overrides = old_frozen_overrides.update(
                        new_frozen_overrides.thaw())
                elif new_frozen_overrides is not None:
                    updated_frozen_overrides = new_frozen_overrides
                elif old_frozen_overrides is not None:
                    updated_frozen_overrides = old_frozen_overrides
                else:
                    # Case: both None
                    updated_frozen_overrides = FrozenOverrides()

                new_steps[updated_frozen_overrides] = old_steps + [new_step]

        self._steps = new_steps
        return self

    def append_step(self, step: PipelineStep) -> 'PipelineBuilder':
        '''Add a single step to every step list in self._steps.

        (used for steps that don't match any selectors because they have no hyperparam overrides)
        '''
        if self._steps is None:
            self._steps = {None: [step]}
            return self
        for key in self._steps:
            self._steps[key].append(step)
        return self

    def build(self) -> Dict[Designator, BoundPipeline]:
        '''Render the final BoundPipelines from lists of steps.'''
        retval: Dict[Designator, BoundPipeline] = {}
        if self._steps is None:
            return retval
        for frozen_overrides, rev_steps in self._steps.items():
            new_pipeline = BoundPipeline.build(
                steps=list(reversed(rev_steps)),
                template_name=self._pipename,
                frozen_overrides=frozen_overrides)
            retval[new_pipeline.designator] = new_pipeline
        return retval


class SearcherImpl(Searcher):
    '''This is the default searcher.'''

    def bind_hyperparams(self, pipeline: BoundPipeline) -> Dict[Designator, BoundPipeline]:
        '''Bind this Searcher's hyperparam range(s) to the pipeline'''
        builder = PipelineBuilder(pipeline.name)
        for step in reversed(pipeline.steps):
            default_alg_ranges = ParamRanges(
                {k: p for k, p in step.hyperparams().items() if isinstance(p, ParamRange)}
            )
            overrides = Overrides(self._overrides)

            no_show: Set[str] = set()  # These are overrides that should not be used in names.
            params = default_alg_ranges.copy()
            for override in overrides:
                if override.selector.matches(step, pipeline):
                    logging.info('Matched: %s\n\tto step: %s\n\tin pipeline: %s\n',
                                 override,
                                 step.pipeline_designator_component,
                                 pipeline.designator)
                    params.update(override.params)
                if override.no_show:
                    no_show.update(override.params.keys())

            if len(params) > 0:
                step_set = StepSet(step)
                for hyperparam_name, param_range in params.items():
                    range_to_multiply = param_range
                    if self._disable_grid_search:
                        # Replace existing range with fixed range at default
                        range_to_multiply = ParamRange(
                            method=RangeMethod.FIXED,
                            prange=param_range.default)
                    step_set.multiply(hyperparam_name=hyperparam_name,
                                      param_range=range_to_multiply,
                                      no_show=no_show)

                builder.append(step_set)
            else:
                builder.append_step(step)

        return builder.build()



'''Test the generator module.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..algorithms.connect import ConnectorModel
from ..algorithms.impl.algorithm import AlgorithmCatalogStub
from ..algorithms.impl.algorithm_auto import AlgorithmCatalogAuto
from ..problem_def.problem_def import ProblemDefinition
from ..templates.impl.pipeline_template import PipelineTemplate
from .designator import Designator
from .generator import GeneratorImpl, GeneratorStub

# pylint: disable=missing-function-docstring,duplicate-code,protected-access

TEST_PROBLEM_DEF = '''{
    "_comments" : [
        "A json file fully encapsulating the problem definition",
        "Specifically, an exmaple time series forecasting problem."
    ],
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "metrics": {
          "MEAN_ABSOLUTE_ERROR": {}
    }
}
'''


def test_init() -> None:
    dut = GeneratorStub(
        name='test_generator',
        algorithm_catalog=AlgorithmCatalogAuto(),
        problem_definition=ProblemDefinition(TEST_PROBLEM_DEF)
    )
    assert dut is not None
    assert dut._algorithm_catalog.__class__.__name__ == "AlgorithmCatalogAuto"


def test_generate_stubbed() -> None:
    dut = GeneratorStub(
        name='stub',
        problem_definition=ProblemDefinition(TEST_PROBLEM_DEF)
    )
    result = dut.generate_all({'stub': PipelineTemplate(name='stub')})
    assert 1 == len(result)
    assert 'stub' == result[Designator('stub')].name


def test_generate_linear() -> None:
    dut = GeneratorImpl(
        algorithm_catalog=AlgorithmCatalogStub(),
        problem_definition=ProblemDefinition(TEST_PROBLEM_DEF)
    )
    pipeline = PipelineTemplate(name='linear_pipeline')
    pipeline.step(ConnectorModel()).set_name('the_step')
    pipeline.step(ConnectorModel())
    pipeline.step(ConnectorModel())
    got = dut.generate(pipeline)
    assert set(got.keys()) == set([Designator('linear_pipeline')])
    assert len(got[Designator('linear_pipeline')].steps) == 3


def test_generate_query() -> None:
    algorithm_catalog = AlgorithmCatalogStub()
    algorithm_catalog.register(ConnectorModel(b='a'), name='atob', tags={'foo': ['bar', 'frob']})
    algorithm_catalog.register(ConnectorModel(d='c'), name='ctod', tags={'foo': ['quux', 'bar']})
    dut = GeneratorImpl(
        algorithm_catalog=algorithm_catalog,
        problem_definition=ProblemDefinition(TEST_PROBLEM_DEF)
    )
    pipeline = PipelineTemplate(name='query_pipeline', algorithm_catalog=algorithm_catalog)
    pipeline.step(ConnectorModel()).set_name('first_step')
    pipeline.query(foo='bar')
    pipeline.step(ConnectorModel()).set_name('last_step')
    got = dut.generate(pipeline)
    key1 = Designator('query_pipeline@atob')
    key2 = Designator('query_pipeline@ctod')
    assert set(got.keys()) == set([key1, key2])
    assert len(got[key1].steps) == 3
    assert got[key2].steps[1].pipeline_designator_component == 'ctod'


def test_generate_parallel() -> None:
    dut = GeneratorImpl(
        algorithm_catalog=AlgorithmCatalogStub(),
        problem_definition=ProblemDefinition(TEST_PROBLEM_DEF)
    )
    pipeline = PipelineTemplate(name='parallel_pipeline', generator=dut)

    pipeline1 = pipeline.new('upper_pipeline')
    pipeline1.step(ConnectorModel()).set_name('upper_step1')
    pipeline1.step(ConnectorModel())
    pipeline2 = pipeline.new('lower_pipeline')
    pipeline2.step(ConnectorModel()).set_name('lower_step')

    pipeline.parallel(upper=pipeline1, lower=pipeline2).set_name('parallel_step')
    pipeline.step(ConnectorModel()).set_name('last_step')

    got = dut.generate(pipeline)

    key = Designator('parallel_pipeline')
    assert set(got.keys()) == set([key])
    assert len(got[key].steps) == 2


def test_generate_all_sunny_day() -> None:
    algorithm_catalog = AlgorithmCatalogStub()
    algorithm_catalog.register(ConnectorModel(b='a'), name='atob', tags={'foo': ['bar', 'frob']})
    algorithm_catalog.register(ConnectorModel(d='c'), name='ctod', tags={'foo': ['quux', 'bar']})

    dut = GeneratorImpl(
        algorithm_catalog=algorithm_catalog,
        problem_definition=ProblemDefinition(TEST_PROBLEM_DEF)
    )

    template1 = PipelineTemplate(name='linear_pipeline')
    template1.step(ConnectorModel()).set_name('the_step')
    template1.step(ConnectorModel())
    template1.step(ConnectorModel())

    template2 = PipelineTemplate(name='query_pipeline', algorithm_catalog=algorithm_catalog)
    template2.step(ConnectorModel()).set_name('first_step')
    template2.query(foo='bar')
    template2.step(ConnectorModel()).set_name('last_step')

    got = dut.generate_all({template1.name: template1, template2.name: template2})

    key1_1 = Designator('linear_pipeline')
    key2_1 = Designator('query_pipeline@atob')
    key2_2 = Designator('query_pipeline@ctod')

    assert set(got.keys()) == set([key1_1, key2_1, key2_2])



'''Tests for distance.py'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import time

from ..algorithms.impl.neighbor_metadata import NeighborMetadata
from ..algorithms.impl.toy_neighbor_state import ToyNeighborState
from ..config_components.distributed_config import DistributedConfig
from ..deciders.impl.decider import Decider
from ..neighbor_manager.node_id import NodeID

from .impl.decider_auto import DeciderCatalogAuto


def test_catalog() -> None:
    '''Test the decider_stub in the DeciderCatalog class.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'distance': {
                    'threshold': 0.2,  # Default is 0.5.
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('distance')(config=distributed.decider.distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=my_state)
    near_id = NodeID(1)
    below_id = NodeID(2)
    distant_id = NodeID(3)
    neighbors = {
        near_id: NeighborMetadata(my_state=ToyNeighborState(center=0.1)),
        below_id: NeighborMetadata(my_state=ToyNeighborState(center=0.3)),
        distant_id: NeighborMetadata(my_state=ToyNeighborState(center=1.0)),
    }

    assert decider.decide(my_meta=my_meta, neighbor_id=near_id, neighbors=neighbors) is False
    assert decider.decide(my_meta=my_meta, neighbor_id=below_id, neighbors=neighbors) is True
    assert decider.decide(my_meta=my_meta, neighbor_id=distant_id, neighbors=neighbors) is True


def test_geometric_term() -> None:
    '''Test the distance decider with a geometric term.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'distance': {
                    'threshold': 8.0,
                    'geometric_term': 0.5,
                    'min': 0.1,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('distance')(config=distributed.decider.distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=my_state)

    neighbor_id = NodeID(1)
    neighbors = {
        neighbor_id: NeighborMetadata(my_state=ToyNeighborState(center=0.3)),
    }

    assert 'distance' not in my_meta.scratch
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 4.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 2.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 1.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 0.5
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 0.25
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True
    assert my_meta.scratch['distance']['threshold'] == 0.125
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True
    assert my_meta.scratch['distance']['threshold'] == 0.1
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True
    assert my_meta.scratch['distance']['threshold'] == 0.1


def test_linear_term() -> None:
    '''Test the distance decider with a geometric term.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'distance': {
                    'threshold': 4.0,
                    'linear_term': -1.0,
                    'min': 0.1,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('distance')(config=distributed.decider.distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=my_state)

    neighbor_id = NodeID(1)
    neighbors = {
        neighbor_id: NeighborMetadata(my_state=ToyNeighborState(center=0.3)),
    }

    assert 'distance' not in my_meta.scratch
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 3.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 2.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 1.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 0.1
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True
    assert my_meta.scratch['distance']['threshold'] == 0.1
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True


def test_interval_seconds() -> None:
    '''Test the distance decider with a geometric term.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'distance': {
                    'threshold': 4.0,
                    'linear_term': -1.0,
                    'min': 0.1,
                    'interval_seconds': 1.0,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('distance')(config=distributed.decider.distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=my_state)

    neighbor_id = NodeID(1)
    neighbors = {
        neighbor_id: NeighborMetadata(my_state=ToyNeighborState(center=0.3)),
    }

    assert 'distance' not in my_meta.scratch
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 3.0
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 3.0
    time.sleep(3.0)
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
    assert my_meta.scratch['distance']['threshold'] == 0.1
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True
    assert my_meta.scratch['distance']['threshold'] == 0.1
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is True



'''Data structures for hyperparameter manipulation.'''
from typing import Any, Dict, Generator, Iterable, List, NamedTuple, Optional

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..wrangler.constants import Matcher
from .matchers import MatcherFactory
from .param_range import ParamRange


class Selector(Dict[Matcher, Any]):
    '''A single selection clause.  The type of the value depends on the key.
    Matcher.DESIGNATOR: Designator
    Matcher.NAME: str
    Matcher.ALGORITHM: str
    Matcher.TAGS: Dict[str, str]
    '''

    def matches(self, step: Any, pipeline: Any) -> bool:
        '''Does this selector clause match the given step and pipeline?'''

        return all(MatcherFactory().make(matcher, args).matches(step, pipeline)
                   for matcher, args in self.items())

    def __str__(self):
        retval = []
        for matcher, val in self.items():
            retval.append(f'{matcher}={val}')
        return '\n'.join(retval)


class AlwaysSelector(Selector):
    '''This class always selects.'''
    def matches(self, step: Any, pipeline: Any) -> bool:
        return True

    def __str__(self) -> str:
        return 'AlwaysSelector'


class ParamRanges(Dict[str, ParamRange]):
    '''Hold names of hyperparams and how we should bind them.

        Key is a hyperparam name.
    '''
    def __str__(self):
        retval = []
        for name, prange in self.items():
            retval.append(f'{name}={prange}')
        return ','.join(retval)


class Override(NamedTuple):
    '''Holds a Selector and its matching Params.'''
    selector: Selector
    params: ParamRanges
    # This override should not be mentioned in designators.
    no_show: bool = False

    # TODO(Merritt): Consider a classmethod to simplify simpler
    # use cases, e.g. Override.fixed_param(selector, name, value)

    def __str__(self) -> str:
        return f'Override({self.selector}, {self.params})'


class Overrides():
    '''Holds a set of overrides.

    Grow Overrides with append or extend.
    Read Overrides by iterating over it.

    This is a single data structure so that it is easier to do more
    intelligent searches of hyperparameter overrides.
    '''
    _value: List[Override]

    def __init__(self, value: Optional[Iterable[Override]] = None):
        self._value = []
        if value is not None:
            self._value = list(value)

    def __iter__(self) -> Generator[Override, None, None]:
        yield from self._value

    def __len__(self) -> int:
        return len(self._value)

    def copy(self) -> 'Overrides':
        '''Return a shallow copy of this Overrides.'''
        return Overrides(self._value)

    def append(self, value: Override):
        '''Add one Override.'''
        self._value.append(value)

    def extend(self, values: Iterable[Override]):
        '''Add a set of Overrides.'''
        self._value.extend(values)

    def __str__(self) -> str:
        return '[' + ', '.join([str(v) for v in self._value]) + ']'



'''Autoloader for the cross-validator catalog.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path

from ...catalog.plugin_catalog import PluginCatalog
from .cross_validator import CrossValidatorCatalog, CrossValidator


class CrossValidatorCatalogAuto(PluginCatalog[CrossValidator], CrossValidatorCatalog):
    '''CrossValidator catalog that automatically loads all cross-validators from a directory.

    Pipeline cross-validators are in ../cross_validators.
    '''

    def __init__(self, **kwargs):
        super().__init__(catalog_name='cross_validators',
                         default_root=Path(__file__).parents[2],
                         **kwargs)



'''Build a virtual environment and install a module in it.

Based loosely on https://gist.github.com/mpurdon/be7f88ee4707f161215187f41c3077f6
'''

from pathlib import Path
from shutil import rmtree
import subprocess
import sys


class VirtualModule:
    '''Create a virtual environment to install a package in.'''
    def __init__(self, virtual_dir: Path):
        self.virtual_dir = virtual_dir
        self.virtual_python = self.virtual_dir / 'bin' / 'python'

    def install_virtual_env(self) -> None:
        '''Install a virtual environment at self.virtual_dir.'''
        subprocess.run([sys.executable, '-m', 'virtualenv', self.virtual_dir], check=True)

    def install_under_venv(self, module: Path) -> None:
        '''Install a module under the virtual environment.'''

        # pypi was flaking several times in a row, so we retry the install up to n-1 times.
        n = 5
        for i in range(n):
            try:
                subprocess.run(
                    [str(self.virtual_python), '-m', 'pip', 'install', str(module)], check=True)
            except subprocess.CalledProcessError:
                if i < n - 1:
                    continue  # Try again.
                raise
            break  # Yay! We're done.

    def finalize(self) -> None:
        '''Clean up the virtual environment.'''
        rmtree(str(self.virtual_dir))



'''Tests for kafka_communicator.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring

# To start a kafka server to run these tests, run the following commands:
# $ docker pull apache/kafka:3.8.0
# $ KAFKA_AUTO_CREATE_TOPICS_ENABLE=TRUE docker run -p 9092:9092 apache/kafka:3.8.0

from copy import deepcopy
import datetime
import getpass
import os
from pathlib import Path
from queue import Queue
import time
from typing import Any, Dict, List, Optional, Set

import pytest

from kafka import KafkaProducer, KafkaConsumer  # type: ignore[import,attr-defined]

from ..config_components.distributed_config import DistributedConfig
from ..neighbor_manager.event import NewNeighbor, Recv
from ..neighbor_manager.node_id import NodeID

from .kafka_communicator import KafkaCommunicator
from .impl.communicator_auto import CommunicatorCatalogAuto

# pylint: disable=duplicate-code, redefined-outer-name
_ = CommunicatorCatalogAuto()


BOOTSTRAP_SERVERS = os.environ.get('BOOTSTRAP_SERVERS', 'localhost:9092')

SHORT_QUEUE_TIMEOUT = 0.5
LONG_QUEUE_TIMEOUT = 30.0


def topic_name(topic: str) -> str:
    '''Generate a unique topic name for the given test.'''
    return f'{Path(__file__).stem}_{getpass.getuser()}_{topic}'


class Producer():
    '''A Kafka producer that sends messages to a topic.

    Only used by distributed algorithms.
    '''
    _producer: KafkaProducer

    def __init__(self):
        self._producer = KafkaProducer(
            bootstrap_servers=BOOTSTRAP_SERVERS.split(','),
            client_id='1')

    def __call__(
        self,
    ) -> KafkaProducer:
        return self._producer


@pytest.fixture(scope='session')
def kafka_producer() -> Producer:
    '''A fixture that returns a Kafka producer.'''
    return Producer()


RECV_TOPIC = topic_name('test_receive')


def test_receive(kafka_producer) -> None:
    '''KafkaCommunicator listens on a topic and receives a message.'''

    known_neighbors: Set[NodeID] = set()
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'kafka',
            'kafka': {
                'bootstrap_servers': BOOTSTRAP_SERVERS.split(','),
                'topic': RECV_TOPIC
            }
        }
    })
    dut = KafkaCommunicator(
        my_id=NodeID(0),
        distributed=distributed,
        known_neighbors=known_neighbors
    )
    queue: Queue = Queue()
    dut.start(queue=queue, timeout=SHORT_QUEUE_TIMEOUT)
    kafka_producer().send(RECV_TOPIC, key=str(NodeID(1)).encode('utf-8'), value=b'Hello, world!')
    try:
        first_event = queue.get(block=True, timeout=LONG_QUEUE_TIMEOUT)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(1)
        assert known_neighbors == {NodeID(1)}

        second_event = queue.get(block=True, timeout=SHORT_QUEUE_TIMEOUT)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(1)
        assert second_event.payload == b'Hello, world!'
    finally:
        dut.stop()


SEND_TOPIC = topic_name('test_send')
SEND_GROUP_ID = 'test_send_group_id'


def test_send() -> None:
    '''KafkaCommunicator talks on node 1, node 0 listens on a topic'''
    known_neighbors: Set[NodeID] = set()
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'kafka',
            'kafka': {
                'bootstrap_servers': BOOTSTRAP_SERVERS.split(','),
                'topic': SEND_TOPIC
            }
        }
    })

    dut = KafkaCommunicator(
        my_id=NodeID(1),
        known_neighbors=known_neighbors,
        distributed=distributed
    )
    queue: Queue = Queue()
    dut.start(queue=queue)

    listener = KafkaConsumer(
        SEND_TOPIC,
        group_id=SEND_GROUP_ID,
        bootstrap_servers=BOOTSTRAP_SERVERS.split(','),
        auto_offset_reset='earliest',
        enable_auto_commit=False
    )

    try:
        dut.send_all(payload=b'Hello World!')
        message = next(listener)
        listener.commit()
        assert message.topic == SEND_TOPIC
        assert message.key == b'1'
        assert message.value == b'Hello World!'
    finally:
        dut.stop()
        listener.close()


EXCHANGE_TOPIC = topic_name('test_exchange')


def test_exchange() -> None:
    '''2 KafkaCommunicator on different nodes talk and listen to each other.'''

    known_neighbors0: Set[NodeID] = set()
    known_neighbors1: Set[NodeID] = set()

    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'kafka',
            'kafka': {
                'bootstrap_servers': BOOTSTRAP_SERVERS.split(','),
                'topic': EXCHANGE_TOPIC,
            }
        }
    })

    dut0 = KafkaCommunicator(
        my_id=NodeID(0),
        known_neighbors=known_neighbors0,
        distributed=distributed
    )
    dut1 = KafkaCommunicator(
        my_id=NodeID(1),
        known_neighbors=known_neighbors1,
        distributed=distributed
    )

    queue0: Queue = Queue()
    queue1: Queue = Queue()

    dut0.start(queue=queue0)
    dut1.start(queue=queue1)

    try:
        payload0 = f'Hello, Node 1! {datetime.datetime.now()}'.encode()
        dut0.send_all(payload=payload0)
        first_event = queue1.get(block=True, timeout=LONG_QUEUE_TIMEOUT)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(0)
        assert known_neighbors1 == {NodeID(0)}

        second_event = queue1.get(block=True, timeout=LONG_QUEUE_TIMEOUT)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(0)
        assert second_event.payload == payload0

        payload1 = f'Nice to meet you, Node 0 :) {datetime.datetime.now()}'.encode()
        dut1.send_all(payload1)

        first_event = queue0.get(block=True, timeout=LONG_QUEUE_TIMEOUT)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(1)
        assert known_neighbors0 == {NodeID(1)}

        second_event = queue0.get(block=True, timeout=LONG_QUEUE_TIMEOUT)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(1)
        assert second_event.payload == payload1

    finally:
        dut0.stop()
        dut1.stop()


SENDALL_TOPIC = topic_name('test_sendall')


def test_sendall() -> None:
    '''One node sends to a bunch of neighbors.'''

    known_neighbors: Dict[NodeID, Set[NodeID]] = {}
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'kafka',
            'kafka': {
                'bootstrap_servers': BOOTSTRAP_SERVERS.split(','),
                'topic': SENDALL_TOPIC,
            }
        }
    })

    dut: Dict[NodeID, KafkaCommunicator] = {}
    queues: Dict[NodeID, Queue] = {}
    for node_id in range(0, 5):
        node = NodeID(node_id)
        known_neighbors[node] = set()
        queues[node] = Queue()
        dut[node] = KafkaCommunicator(
            my_id=node,
            known_neighbors=known_neighbors[node],
            distributed=distributed)
        dut[node].start(queue=queues[node], timeout=SHORT_QUEUE_TIMEOUT)

    # For KafkaCommunicator, there is no need for neighbors to introduce themselves.
    try:
        # This is what we are testing.
        dut[NodeID(0)].send_all(payload=b'Hello, All Nodes!')

        # Confirm that the test worked.
        for node_id in range(1, 5):
            node = NodeID(node_id)

            first_event = queues[node].get(block=True, timeout=LONG_QUEUE_TIMEOUT)
            assert isinstance(first_event, NewNeighbor)
            assert first_event.neighbor == NodeID(0)
            assert known_neighbors[node] == {NodeID(0)}

            second_event = queues[node].get(block=True, timeout=LONG_QUEUE_TIMEOUT)
            assert isinstance(second_event, Recv)
            assert second_event.neighbor == NodeID(0)
            assert known_neighbors[node] == {NodeID(0)}
            assert second_event.payload == b'Hello, All Nodes!'

    finally:
        for node_id in range(0, 5):
            node = NodeID(node_id)
            dut[node].stop()


DYNAMIC_DISCOVERY_TOPIC = topic_name('test_dynamic_discovery')


def test_dynamic_discovery() -> None:
    '''A bunch of neighbors send to one node who learns their identities.'''

    known_neighbors: Dict[NodeID, Set[NodeID]] = {}
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'kafka',
            'kafka': {
                'bootstrap_servers': BOOTSTRAP_SERVERS.split(','),
                'topic': DYNAMIC_DISCOVERY_TOPIC,
            }
        }
    })

    dut: Dict[NodeID, KafkaCommunicator] = {}
    queues: Dict[NodeID, Queue] = {}
    for node_id in range(0, 5):
        node = NodeID(node_id)
        known_neighbors[node] = set()
        queues[node] = Queue()
        dut[node] = KafkaCommunicator(
            my_id=node,
            known_neighbors=known_neighbors[node],
            distributed=distributed)
        dut[node].start(queue=queues[node], timeout=SHORT_QUEUE_TIMEOUT)

    # This is the test. Everybody introduces themselves.
    for node_id in range(1, 5):
        node = NodeID(node_id)
        dut[node].send_all(f'Hello, I am {node}.'.encode())

    # Confirm that the test worked.
    try:
        # Read a NewNeighbor and a Recv for each neighbor.
        # The pairs of requests are randomly interleaved due to threading,
        # so we don't bother checking what they are.
        # This is needed to make sure every message has been recieved before
        #   checking known_neighbors.
        # We receive a NewNeighbor and a Recv from each neighbor.
        for _ in range(1, 5):
            _ = queues[NodeID(0)].get()
            _ = queues[NodeID(0)].get()

        assert known_neighbors[NodeID(0)] == {
            NodeID(1),
            NodeID(2),
            NodeID(3),
            NodeID(4),
        }

    finally:
        for node_id in range(0, 5):
            node = NodeID(node_id)
            dut[node].stop()


KAFKA_TOPIC = topic_name('test_kafka')


@pytest.mark.skip(
    reason='This test deliberately fails in order to provide exploratory information.')
def test_kafka() -> None:
    '''Test Kafka.'''
    producer = KafkaProducer(
        bootstrap_servers=BOOTSTRAP_SERVERS.split(','),
        client_id='kafka_producer_client_id')

    consumer = KafkaConsumer(
        KAFKA_TOPIC,
        group_id='kafka_consumer_group_id',
        client_id='kafka_consumer_client_id',
        bootstrap_servers=BOOTSTRAP_SERVERS.split(','),
        auto_offset_reset='latest',
        enable_auto_commit=True)

    # send 2 messages with producer that log time that they were sent

    producer.send(KAFKA_TOPIC,
                  key=b'some key',
                  value=f'Message 1: {datetime.datetime.now()}'.encode('utf-8'))
    time.sleep(0.5)
    producer.send(KAFKA_TOPIC,
                  key=b'some key',
                  value=f'Message 2: {datetime.datetime.now()}'.encode('utf-8'))

    # print all the messages in consumer
    # topics = consumer.poll(timeout_ms=2000)  # Dict[TopicPartition, List[ConsumerRecord]]
    # print('topics: %s' % topics)
    # for topic in topics:
    #     print(f"DEBUG: topic: {topic}")
    #     for message in topics[topic]:
    for message in consumer:
        # message value and key are raw bytes -- decode if necessary!
        # e.g., for unicode: `message.value.decode('utf-8')`
        print(f'{message.topic}:{message.partition}:{message.offset}: '
              f'key={str(message.key)} value={str(message.value)}')
    assert False


SEND_UNICAST_PREFIX = topic_name('test_send_unicast')
SEND_UNICAST_GROUP_ID = 'test_send_unicast_group_id'


def test_send_unicast() -> None:
    '''KafkaCommunicator talks on node 0, nodes 1 and 2 listen on their topics.'''
    known_neighbors: Set[NodeID] = set()
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'kafka',
            'strategy': 'unicast',
            'kafka': {
                'bootstrap_servers': BOOTSTRAP_SERVERS.split(','),
                'topic_prefix': SEND_UNICAST_PREFIX,
            }
        },
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '0': [1, 2],
                    '1': [0, 2],
                    '2': [0, 1],
                }
            }
        },
        'my_id': 0,
    })

    dut = KafkaCommunicator(
        my_id=NodeID(1),
        known_neighbors=known_neighbors,
        distributed=distributed
    )
    queue: Queue = Queue()
    dut.start(queue=queue)

    listener1 = KafkaConsumer(
        f'{SEND_UNICAST_PREFIX}1',
        group_id=SEND_UNICAST_GROUP_ID,
        bootstrap_servers=BOOTSTRAP_SERVERS.split(','),
        auto_offset_reset='earliest',
        enable_auto_commit=False
    )

    listener2 = KafkaConsumer(
        f'{SEND_UNICAST_PREFIX}2',
        group_id=SEND_UNICAST_GROUP_ID,
        bootstrap_servers=BOOTSTRAP_SERVERS.split(','),
        auto_offset_reset='earliest',
        enable_auto_commit=False
    )

    try:
        dut.send(dest_ids=[NodeID(1), NodeID(2)], payload=b'Hello World!')

        # TODO(piggy): Figure out why the second next(listenern) is hanging.
        # message1 = next(listener1)
        # listener1.commit()
        # assert message1.topic == f'{SEND_UNICAST_PREFIX}1'
        # assert message1.key == b'1'
        # assert message1.value == b'Hello World!'

        message2 = next(listener2)
        listener2.commit()
        assert message2.topic == f'{SEND_UNICAST_PREFIX}2'
        assert message2.key == b'1'
        assert message2.value == b'Hello World!'

    finally:
        dut.stop()
        listener1.close()
        listener2.close()


THREE_NODE_PREFIX = topic_name('test_3_node_unicast')


def test_3_node_unicast() -> None:
    '''3 KafkaCommunicators on different nodes talk to and listen to each other.'''

    distributed: Dict[str, Any] = {
        'communicator': {
            'name': 'kafka',
            'kafka': {
                'bootstrap_servers': BOOTSTRAP_SERVERS.split(','),
                'topic_prefix': THREE_NODE_PREFIX,
            },
            'strategy': 'unicast',
        },
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '0': [1, 2],
                    '1': [0, 2],
                    '2': [0, 1],
                }
            }
        },
        'my_id': None,
    }

    known_neighbors: List[Set[NodeID]] = [set(), set(), set()]
    duts: List[Optional[KafkaCommunicator]] = [None, None, None]
    configs: List[Dict[str, Any]] = [{}, {}, {}]
    queues: List[Queue] = [Queue(), Queue(), Queue()]

    for i in range(3):
        configs[i] = deepcopy(distributed)
        configs[i]['my_id'] = i
        duts[i] = KafkaCommunicator(
            my_id=NodeID(i),
            known_neighbors=known_neighbors[i],
            distributed=DistributedConfig(clause=configs[i])
        )

    for i in range(3):
        dut = duts[i]
        assert dut is not None
        dut.start(queue=queues[i])

    try:
        payload0 = f'Hello, Node 1! {datetime.datetime.now()}'.encode()
        dut = duts[0]
        assert dut is not None
        dut.send(dest_ids=[NodeID(1)], payload=payload0)
        first_event = queues[1].get(block=True, timeout=LONG_QUEUE_TIMEOUT)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(0)
        assert known_neighbors[1] == {NodeID(0)}

        second_event = queues[1].get(block=True, timeout=LONG_QUEUE_TIMEOUT)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(0)
        assert second_event.payload == payload0

        payload1 = f'Hello nodes 2 and 0 {datetime.datetime.now()}'.encode()
        dut = duts[1]
        assert dut is not None
        dut.send(dest_ids=[NodeID(0), NodeID(2)], payload=payload1)

        for i in [0, 2]:
            first_event = queues[i].get(block=True, timeout=LONG_QUEUE_TIMEOUT)
            assert isinstance(first_event, NewNeighbor)
            assert first_event.neighbor == NodeID(1)
            assert known_neighbors[i] == {NodeID(1)}

            second_event = queues[i].get(block=True, timeout=LONG_QUEUE_TIMEOUT)
            assert isinstance(second_event, Recv)
            assert second_event.neighbor == NodeID(1)
            assert second_event.payload == payload1

    finally:
        for i in range(3):
            if duts[i] is not None:
                dut = duts[i]
                assert dut is not None
                dut.stop()



'''Holds configuration information for aggregation.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..config_components.impl.config_component import ConfigComponent, ConfigError


class AggregationError(ConfigError):
    '''Accummulating error for rank aggregation parsing.'''

    errors: List[ConfigError]

    def __init__(self, errors: List[ConfigError]):
        super().__init__(f'At least one validation error occured: {errors!r}')
        self.errors = errors


class AggregationConfig(ConfigComponent):
    '''Holds configuration for rank aggregation.

    If this clause is missing, we do not do rank aggregation.
    '''
    _method: List[str]

    class Keys(AEnum):
        '''Top-level clauses in aggregation component'''
        METHOD = 'method'

    def __init__(self, clause: Dict[str, Any]):
        super().__init__(clause=clause)
        method = self._get_with_default(self.Keys.METHOD, dflt=[])
        if method is not None:
            if not isinstance(method, list):
                assert isinstance(method, str)
                method = [method]
        self._method = method
        self.validate()

    @property
    def method(self) -> List[str]:
        '''Returns the aggregation method(s).'''
        return self._method

    def validate(self, **kwargs) -> None:
        pass



'''An instance of a fittable algorithm that follows sklearn conventions.'''
import pickle
from typing import Any, Optional

from ...wrangler.dataset import Dataset
from ...wrangler.dataset import DatasetKeys
from .algorithm import Algorithm
from .algorithm_instance import DatasetError
from .fittable_algorithm_instance import FittableAlgorithmInstance, UntrainedError


class SklearnLikeAlgorithmInstance(FittableAlgorithmInstance):
    '''An instance of a fittable algorithm that follows sklearn conventions.'''
    _impl: Any
    _constructor: type

    def __init__(self, parent: Algorithm, *args, **kwargs):
        super().__init__(parent, *args)
        self._impl = self._constructor(**self.algorithm.hyperparams(**kwargs))

    def deserialize(self, serialized_model: bytes) -> 'SklearnLikeAlgorithmInstance':
        '''Restore the serialized version of a model.'''
        self._impl = pickle.loads(serialized_model)
        self._trained = True
        return self

    def serialize(self) -> bytes:
        '''Return a serialized version of a trained model.'''
        if not self._trained:
            raise UntrainedError(f'attempt to save training before fit for {self.algorithm.name}')

        return pickle.dumps(self._impl, pickle.HIGHEST_PROTOCOL)

    def fit(self, dataset: Optional[Dataset]) -> None:
        '''Fit a model based on input dataset.'''
        if dataset is None:
            raise DatasetError(f'attempt to fit with no data for {self.algorithm.name}')
        try:
            self._impl.fit(dataset[DatasetKeys.COVARIATES.value], dataset[DatasetKeys.TARGET.value])
        except KeyError as err:
            raise DatasetError(f'fit dataset malformed {dataset!r}') from err
        self._trained = True

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if not self._trained:
            raise UntrainedError(f'attempt to predict before fit for {self.algorithm.name}')
        if dataset is None:
            return None
        try:
            result = self._impl.predict(dataset[DatasetKeys.COVARIATES.value])
        except KeyError as err:
            raise DatasetError(f'predict dataset malformed {dataset!r}') from err
        retval = dataset.output()
        retval.predictions = result
        return retval



'''Contains information on hyperparam overrides to use for indexing dictionaries.'''
from collections import defaultdict
from typing import Dict

from frozendict import frozendict


class FrozenOverrides(frozendict[str, frozendict[str, str]]):
    '''Contains information on hyperparam overrides to use in building Designator

    The contents look like this\\:

    .. code-block:: Python

        {
            'pipeline_designator_component': {
                'hyperparam_name': 'hyperparam_value_str'
            }
        }

    '''

    @classmethod
    def freeze(cls, inp: Dict[str, Dict[str, str]]) -> 'FrozenOverrides':
        '''Turn a dict of dicts into a deeply immutable OverridesInfo'''
        cool: Dict[str, frozendict[str, str]] = {
            k: frozendict(v) for k, v in inp.items()}
        return FrozenOverrides(cool)

    def thaw(self) -> Dict[str, Dict[str, str]]:
        '''Turn self into a deeply mutable dict of dicts.'''
        retval: Dict[str, Dict[str, str]] = defaultdict(dict)
        for step_des, hyperparams in self.items():
            retval[step_des].update(hyperparams)
        return retval

    def update(self, other: Dict[str, Dict[str, str]]) -> 'FrozenOverrides':
        '''Update self with values in other'''
        tmp = self.thaw()
        for step_des, hyperparams in tmp.items():
            if step_des in other:
                hyperparams.update(other[step_des])
        for step_des, hyperparams in other.items():
            if step_des not in tmp:
                tmp[step_des] = hyperparams
        return self.__class__.freeze(tmp)
